"""
  Unit tests for the genetic algorithm optimiser.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=abstract-class-little-used

from copy import deepcopy
from argparse import Namespace
import unittest
# Local
from . import nn_ga_optimiser
from .nn_domains import NNDomain
from .syn_nn_functions import cnn_syn_func1, mlp_syn_func1
from .nn_domains import CNNConstraintChecker, MLPConstraintChecker
from .nn_modifiers import get_nn_modifier_from_args
from .unittest_nn_modifier_class import networks_are_equal
from .nn_opt_utils import get_initial_cnn_pool, get_initial_mlp_pool
from ..exd.experiment_caller import FunctionCaller
from ..exd.worker_manager import SyntheticWorkerManager
from ..utils.reporters import get_reporter
from ..utils.option_handler import load_options
from ..utils.base_test_class import BaseTestClass, execute_tests


def check_if_pre_eval_networks_have_changed(options_1, options_2):
  """ Checks if the pre_eval networks have changed. Raises an error if they have. """
  assert options_1.pre_eval_vals is not options_2.pre_eval_vals and \
         options_1.pre_eval_vals == options_2.pre_eval_vals
  assert options_1.pre_eval_true_vals is not options_2.pre_eval_true_vals and \
         options_1.pre_eval_true_vals == options_2.pre_eval_true_vals
  for idx in range(len(options_1.pre_eval_points)):
    assert options_1.pre_eval_points[idx] is not options_2.pre_eval_points[idx]
    assert networks_are_equal(options_1.pre_eval_points[idx],
                              options_2.pre_eval_points[idx])

def get_optimiser_args(tester, nn_type, optimiser_args):
  """ Returns arguments for the optimiser. """
  reporter = get_reporter('default')
  options = load_options(optimiser_args, reporter=reporter)
  if nn_type == 'cnn':
    options.pre_eval_points = tester.cnn_init_pool
    options.pre_eval_vals = tester.cnn_init_vals
    options.pre_eval_true_vals = tester.cnn_init_vals
    constraint_checker = tester.cnn_constraint_checker
    mutation_op = tester.cnn_mutation_op
  else:
    options.pre_eval_points = tester.mlp_init_pool
    options.pre_eval_vals = tester.mlp_init_vals
    options.pre_eval_true_vals = tester.mlp_init_vals
    constraint_checker = tester.mlp_constraint_checker
    mutation_op = tester.mlp_mutation_op
  options_clone = deepcopy(options)
  return options, options_clone, reporter, constraint_checker, mutation_op

def get_nn_opt_arguments():
  """ Returns arguments for NN Optimisation. """
  ret = Namespace()
  ret.cnn_constraint_checker = CNNConstraintChecker(50, 5, 1e8, 0, 5, 5, 200, 1024, 8)
  ret.mlp_constraint_checker = MLPConstraintChecker(50, 5, 1e8, 0, 5, 5, 200, 1024, 8)
  ret.cnn_mutation_op = get_nn_modifier_from_args(ret.cnn_constraint_checker,
                                                   [0.5, 0.25, 0.125, 0.075, 0.05])
  ret.mlp_mutation_op = get_nn_modifier_from_args(ret.mlp_constraint_checker,
                                                   [0.5, 0.25, 0.125, 0.075, 0.05])
  # Create the initial pool
  ret.cnn_init_pool = get_initial_cnn_pool()
  ret.cnn_init_vals = [cnn_syn_func1(cnn) for cnn in ret.cnn_init_pool]
  ret.mlp_init_pool = get_initial_mlp_pool('reg')
  ret.mlp_init_vals = [mlp_syn_func1(mlp) for mlp in ret.mlp_init_pool]
  # Create a domain
  ret.nn_domain = NNDomain('cnn', ret.cnn_constraint_checker)
  ret.cnn_domain = NNDomain('cnn', ret.cnn_constraint_checker)
  ret.mlp_domain = NNDomain('mlp-reg', ret.mlp_constraint_checker)
  # Return
  return ret


# Tester class ========================================================================
@unittest.skip
class GAOptimiserTestCase(BaseTestClass):
  """ Unit test for GA Optimisation. """

  def setUp(self):
    """ Set up. """
    ret = get_nn_opt_arguments()
    for key, val in ret.__dict__.items():
      setattr(self, key, val)

  def test_instantiation(self):
    """ Test creation of object. """
    self.report('Testing GA Optimiser instantiation.')
    func_caller = FunctionCaller(cnn_syn_func1, self.nn_domain)
    worker_manager = SyntheticWorkerManager(1, time_distro='const')
    optimiser = nn_ga_optimiser.NNGAOptimiser(func_caller,
      worker_manager, self.cnn_mutation_op, reporter='silent')
    self.report('Instantiated GAOptimiser object.')
    for attr in dir(optimiser):
      if not attr.startswith('_'):
        self.report('optimiser.%s = %s'%(attr, str(getattr(optimiser, attr))),
                    'test_result')

  def _get_optimiser_args(self, nn_type):
    """ Returns the options and reporter. """
    return get_optimiser_args(self, nn_type, nn_ga_optimiser.nn_ga_opt_args)

  @classmethod
  def _test_optimiser_results(cls, opt_val, _, history, options, options_clone):
    """ Tests optimiser results. """
    assert opt_val == history.curr_opt_vals[-1]
    check_if_pre_eval_networks_have_changed(options, options_clone)

  def test_ga_optimisation_single(self):
    """ Test optimisation. """
    self.report('Testing GA Optimiser with just one worker.')
    worker_manager = SyntheticWorkerManager(1, time_distro='const')
    func_caller = FunctionCaller(cnn_syn_func1, self.nn_domain)
    options, options_clone, reporter, _, mutation_op = self._get_optimiser_args('cnn')
    opt_val, opt_pt, history = nn_ga_optimiser.nn_ga_optimise_from_args(
      func_caller, worker_manager, 20, 'asy', mutation_op,
      options=options, reporter=reporter)
    self._test_optimiser_results(opt_val, opt_pt, history, options, options_clone)
    self.report('')

  def test_optimisation_asynchronous(self):
    """ Test optimisation. """
    self.report('Testing GA Optimiser with four workers asynchronously.')
    worker_manager = SyntheticWorkerManager(4, time_distro='halfnormal')
    func_caller = FunctionCaller(mlp_syn_func1, self.nn_domain)
    options, options_clone, reporter, _, mutation_op = self._get_optimiser_args('mlp')
    opt_val, opt_pt, history = nn_ga_optimiser.nn_ga_optimise_from_args(
      func_caller, worker_manager, 20, 'asy', mutation_op,
      options=options, reporter=reporter)
    self._test_optimiser_results(opt_val, opt_pt, history, options, options_clone)
    self.report('')


if __name__ == '__main__':
  execute_tests()




"""
  Multi-Objective Gaussian Process Bandit Optimisation.
  - bparia@cs.cmu.edu
  - kandasamy@cs.cmu.edu

  The framework allows implementing several methods for multi-objective Bayesian
  optimisation, although we have only implemented the MOORS algorithm from,
  "A Flexible Framework for Multi-Objective Bayesian Optimization using Random
   Scalarizations", Paria, Kandasamy, Poczos.
  (Available at https://arxiv.org/pdf/1805.12168.pdf.)
"""

from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=redefined-builtin

from argparse import Namespace
import numpy as np

# Local imports
from ..exd.cp_domain_utils import get_processed_func_from_raw_func_for_cp_domain, \
                                load_cp_domain_from_config_file
from ..exd import domains
from ..exd.exd_utils import get_euclidean_initial_qinfos, get_cp_domain_initial_qinfos
from ..exd.experiment_caller import CPMultiFunctionCaller
from ..gp.euclidean_gp import EuclideanGPFitter
from ..gp.cartesian_product_gp import CPGPFitter
from . import multiobjective_gpb_acquisitions
from .blackbox_optimiser import CalledMFOptimiserWithSFCaller
from .multiobjective_optimiser import MultiObjectiveOptimiser
from .gp_bandit import GPBandit, get_all_euc_gp_bandit_args, get_all_cp_gp_bandit_args
from ..utils.option_handler import get_option_specs, load_options
from ..utils.reporters import get_reporter


_NO_MF_FOR_MOGPB_ERR_MSG = 'Multi-fidelity support has not been implemented yet' + \
                           ' for multi-objective GP Bandits.'

multiobjective_gp_bandit_args = [
  get_option_specs('moo_strategy', False, 'moors', \
    'Get name of multi-objective strategy. So far, Dragonfly only supports moors.'),
  # Arguments for MOORS
  get_option_specs('moors_scalarisation', False, 'tchebychev', \
    'Scalarisation for MOORS. Should be "tchebychev" or "linear".'),
  get_option_specs('moors_weight_sampler', False, 'flat_uniform', \
    'A weight sampler for moors.'),
  get_option_specs('moors_reference_point', False, None, \
    'Reference point for MOORS.'),
]


def get_all_euc_moo_gp_bandit_args(additional_args=None):
  """ Returns all arguments for Euclidean MOO GP Bandits. """
  if additional_args is None:
    additional_args = []
  return get_all_euc_gp_bandit_args(additional_args + multiobjective_gp_bandit_args)

def get_all_cp_moo_gp_bandit_args(additional_args=None):
  """ Returns all arguments for Euclidean MOO GP Bandits. """
  if additional_args is None:
    additional_args = []
  return get_all_cp_gp_bandit_args(additional_args + multiobjective_gp_bandit_args)


# Some utilities for MOORS

def _get_moors_weight_sampler(multi_func_caller, weight_sampler):
  """ Returns a weight sampler. """
  if hasattr(weight_sampler, '__call__'):
    return weight_sampler
  elif weight_sampler == 'flat_uniform':
    def _get_uniform_weight_sampler(_num_funcs):
      """ Draws flat uniform weights. """
      return lambda: np.abs(np.random.normal(loc=0.0, scale=10, size=(_num_funcs,)))
    return _get_uniform_weight_sampler(multi_func_caller.num_funcs)
  else:
    raise ValueError('Could not process argument for weight_sampler: %s.'%(
                      weight_sampler))

def _get_moors_reference_point(multi_func_caller, reference_point):
  """ Returns the reference point. """
  if hasattr(reference_point, '__len__') and \
    len(reference_point) == multi_func_caller.num_funcs:
    return reference_point
  elif reference_point is None:
    return [-1.0] * multi_func_caller.num_funcs
  else:
    raise ValueError('Could not process argument for reference_point: %s.'%(
                     reference_point))


def get_default_moo_acquisition_for_domain(domain):
  """ Returns the default acquisition for the domain. """
  if domain.get_type() == 'euclidean':
#     return 'ucb-ts-add_ucb'
    return 'ucb-ts'
  else:
    return 'ucb-ts'


# The MultiObjectiveGPBandit Class
# ========================================================================================
class MultiObjectiveGPBandit(MultiObjectiveOptimiser, GPBandit):
  """ MultiObjGPBandit Class. """
  # pylint: disable=attribute-defined-outside-init

  # Constructor.
  def __init__(self, multi_func_caller, worker_manager, is_mf=False,
               options=None, reporter=None):
    """ Constructor. """
    self._is_mf = is_mf
    if is_mf:
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    if is_mf and not multi_func_caller.is_mf():
      raise CalledMFOptimiserWithSFCaller(self, multi_func_caller)
    super(MultiObjectiveGPBandit, self).__init__(multi_func_caller, worker_manager,
                                                 None, options=options, reporter=reporter)

  def _get_method_str(self):
    """ Returns a string describing the method. """
    gpb_str = 'mfmobo-%s'%(self.options.mf_strategy) if self.is_an_mf_method() else 'mobo'
    return '%s(%s)'%(gpb_str, '-'.join(self.acqs_to_use))

  def _multi_opt_method_set_up(self):
    """ Some set up for the GPBandit class. """
    # Set up acquisition optimisation
    self.gps = [None] * self.multi_func_caller.num_funcs
    super(MultiObjectiveGPBandit, self)._opt_method_set_up()
    delattr(self, 'gp')
    # MOO set up
    if self.options.moo_strategy == 'moors':
      self.moors_weight_sampler = _get_moors_weight_sampler(self.multi_func_caller,
                                                  self.options.moors_weight_sampler)
      self.moors_reference_point = _get_moors_reference_point(self.multi_func_caller,
                                                  self.options.moors_reference_point)
    else:
      raise ValueError('Unknown MOO Strategy %s.'%(self.options.moo_strategy))
    # Do any child specific set up
    self._domain_specific_multi_opt_method_set_up()

  @classmethod
  def _get_default_acquisition_for_domain(cls, domain):
    """ Return default acqusition for domain. """
    return get_default_moo_acquisition_for_domain(domain)

  def _domain_specific_multi_opt_method_set_up(self):
    """ Any domain specific set up for multi-objective GP bandit. """
    pass

  @classmethod
  def _compare_two_sets_of_obj_values(cls, obj_vals_1, obj_vals_2):
    """ Compares two sets of objective values and returns a 3-tuple where the first
        element is the number of objectives on which obj_vals_1 is better than obj_vals_2,
        the third vice versa, and the second is the number of equal objective values.
    """
    ret = [0, 0, 0]
    for obj1, obj2 in zip(obj_vals_1, obj_vals_2):
      if obj1 > obj2:
        ret[0] += 1
      elif obj1 == obj2:
        ret[1] += 1
      else:
        ret[2] += 1
    return tuple(ret)

  def _multi_opt_method_update_history(self, qinfo):
    """ Update history for GP bandit specific statistics. """
    if len(self.history.curr_pareto_vals) >= 2:
      num_improvements, _, _ = self._compare_two_sets_of_obj_values(
        self.history.curr_pareto_vals[-1], self.history.curr_pareto_vals[-2])
      if hasattr(qinfo, 'curr_acq'):
        self.acqs_to_use_counter[qinfo.curr_acq] += 1
        if self.options.acq_probs == 'adaptive':
          self.acq_sampling_weights[qinfo.curr_acq] += num_improvements
      if hasattr(self, 'gp_processors') and hasattr(qinfo, 'hp_tune_method'):
        for gp_proc in self.gp_processors:
          gp_proc.gp_fitter.update_hp_tune_method_weight(qinfo.hp_tune_method,
                                                         num_improvements)
      self._domain_specific_multi_opt_method_update_history(qinfo, num_improvements)

  def _domain_specific_multi_opt_method_update_history(self, qinfo, num_improvements):
    """ Updates to the history specific to the domain. """
    pass

  # Managing the GP ---------------------------------------------------------
  def _set_next_gp(self):
    """ Returns the next GP. """
    if not hasattr(self, 'gp_processors') or self.gp_processors is None:
      self._build_new_gps()
    self.gps = []
    for gp_processor in self.gp_processors:
      ret = gp_processor.gp_fitter.get_next_gp()
      gp_processor.fit_type = ret[0]
      gp_processor.hp_tune_method = ret[1]
      self.gps.append(ret[2])
    self._domain_specific_set_next_gp()
    for i, gp_processor in enumerate(self.gp_processors):
      if gp_processor.fit_type in ['sample_hps_with_probs', \
                                        'post_sample_hps_with_probs']:
        reg_data = self._get_moo_gp_reg_data(i)
        self._child_set_moo_gp_data(reg_data, i)
    if self.step_idx == self.last_model_build_at and \
       self.options.report_model_on_each_build:
      self._report_current_gps()

  def _domain_specific_set_next_gp(self):
    """ Sets the next GP in child class """
    pass

  def _child_set_gp_data(self, reg_data):
    """ Set data in child. """
    raise NotImplementedError('Use _get_moo_gp_reg_data instead!')

  def _child_set_moo_gp_data(self, reg_data, obj_ind):
    """ Set data in child. """
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    else:
      self.gps[obj_ind].set_data(reg_data[0], reg_data[1], build_posterior=True)

  def _child_build_new_model(self):
    """ Builds a new model. """
    self._build_new_gps()

  def _report_model(self):
    """ Report the current model. """
    # We will do this separately since posterior sampling complicates how we keep
    # track of the GPs. See _set_next_gp()
    pass

  def _report_current_gps(self):
    """ Reports the current GP. """
    for gp_idx, gp in enumerate(self.gps):
      gp_fit_report_str = '    -- GP-%d at iter %d: %s'%(
                            gp_idx, self.step_idx, str(gp))
      self.reporter.writeln(gp_fit_report_str)

  def _get_opt_method_header_str(self):
    """ Header for optimisation Method. """
    return ', acqs=<num_times_each_acquisition_was_used>'

  def _get_opt_method_report_results_str(self):
    """ Any details to include in a child method when reporting results.
    """
    ret_list = ['%s:%d'%(key, self.acqs_to_use_counter[key]) for key in self.acqs_to_use]
    ret = 'acqs=[' + ', '.join(ret_list) + ']'
    return ', ' + ret

  def _get_gp_reg_data(self):
    """ Returns the current data to be added to the GP. """
    raise NotImplementedError('Use _get_moo_gp_reg_data instead!')

  def _get_moo_gp_reg_data(self, obj_ind):
    """ Returns the current data to be added to the GP. """
    # pylint: disable=no-member
    reg_X = self.prev_eval_points + self.history.query_points
    reg_Y = self.prev_eval_vals + self.history.query_vals
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    else:
      return reg_X, [y[obj_ind] for y in reg_Y]

  def _get_gp_fitter(self, reg_data, use_additive=False):
    """ Returns a GP Fitter. """
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    else:
      return self._get_non_mf_gp_fitter(reg_data, use_additive)

  def _get_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the Multi-fidelity GP Fitter. Can be overridded by a child class. """
    raise NotImplementedError('Implement in a Child class.')

  def _get_non_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the NOn-Multi-fidelity GP Fitter. Can be overridded by a child class. """
    raise NotImplementedError('Implement in a Child class.')

  def _build_new_gps(self):
    """ Builds a GP with the data in history and stores in self.gp. """
    if hasattr(self.multi_func_caller, 'init_gps') and\
       self.multi_func_caller.init_gps is not None:
      # If you know the true GP.
      raise NotImplementedError('Not implemented passing given GP yet.')
    else:
      if self.options.shrink_kernel_with_time:
        raise NotImplementedError('Not implemented kernel shrinking for the GP yet.')
      # Invoke the GP fitter.
      self.gp_processors = []
      for i in range(self.multi_func_caller.num_funcs):
        reg_data = self._get_moo_gp_reg_data(i)
        gp_fitter = self._get_gp_fitter(reg_data)
        # Fits gp and adds it to gp_processor
        gp_fitter.fit_gp_for_gp_bandit(self.options.build_new_model_every)
        gp_processor = Namespace()
        gp_processor.gp_fitter = gp_fitter
        self.gp_processors.append(gp_processor)
      self.gps = None # Mostly to avoid bugs
      self._domain_specific_build_new_gps()

  def _domain_specific_build_new_gps(self):
    """ Builds an extra GP in child class if required. """
    pass

  def _domain_specific_build_new_gp(self, _):
    """ Builds an extra GP in child class if required. """
    raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)

  def _add_data_to_model(self, qinfos):
    """ Add data to self.gp """
    if self.gps is None:
      return
    if len(qinfos) == 0:
      return
    new_points = [qinfo.point for qinfo in qinfos]
    new_vals = [qinfo.val for qinfo in qinfos]
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    else:
      self._add_data_to_gps((new_points, new_vals))

  def _add_data_to_gps(self, new_data):
    """ Adds data to the GP. """
    # Add data to the GP only if we will be repeating with the same GP.
    if hasattr(self, 'gp_processors') and hasattr(self.gp_processors[0], 'fit_type') and \
      self.gp_processors[0].fit_type == 'fitted_gp':
      if self.is_an_mf_method():
        raise NotImplementedError("MF with MO is not supported yet")
      else:
        for i, gp in enumerate(self.gps):
          if self.gp_processors[i].fit_type == 'fitted_gp':
            vals = [y[i] for y in new_data[1]]
            gp.add_data_multiple(new_data[0], vals)
    self._domain_specific_add_data_to_gp(new_data)

  def _domain_specific_add_data_to_gp(self, new_data):
    """ Adds data from the child class to the GP. """
    pass

  # Methods needed for initialisation ----------------------------------------
  def _child_optimise_initialise(self):
    """ No additional initialisation for GP bandit. """
    self._build_new_gps()

  # Methods needed for optimisation ------------------------------------------
  def _get_next_acq(self):
    """ Gets the acquisition to use in the current iteration. """
    if self.options.acq_probs == 'adaptive':
      self.acq_probs = self._get_adaptive_ensemble_acq_probs()
    ret = np.random.choice(self.acqs_to_use, p=self.acq_probs)
    return ret

  def _get_ancillary_data_for_acquisition(self, curr_acq):
    """ Returns ancillary data for the acquisitions. """
    max_num_acq_opt_evals = self.get_acq_opt_max_evals(self.step_idx)
    ret = Namespace(curr_acq=curr_acq,
                    max_evals=max_num_acq_opt_evals,
                    t=self.step_idx,
                    domain=self.domain,
                    curr_pareto_vals=self.curr_pareto_vals,
                    eval_points_in_progress=self.eval_points_in_progress,
                    acq_opt_method=self.acq_opt_method,
                    handle_parallel=self.options.handle_parallel,
                    mf_strategy=self.options.mf_strategy,
                    is_mf=self.is_an_mf_method(),
                    num_funcs=self.multi_func_caller.num_funcs,
                   )
    if curr_acq == 'add_ucb':
      ret.domain_bounds = self.domain.bounds
    if self.options.moo_strategy == 'moors':
      ret.obj_weights = self.moors_weight_sampler()
      ret.reference_point = self.moors_reference_point
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    return ret

  def _get_name_in_multiobjective_acquisition_module(self, curr_acq):
    """ Returns the name of the acquisition as used in multiobjective_gpb_acquisitions.py.
    """
    if self.options.moo_strategy == 'moors':
      scalarisation_str = 'lin' if self.options.moors_scalarisation == 'linear' else 'tch'
      return scalarisation_str + '_' + curr_acq
    else:
      raise ValueError('Unknown MOO Strategy %s.'%(self.options.moo_strategy))

  def _determine_next_query(self):
    """ Determine the next point for evaluation. """
    curr_acq = self._get_next_acq()
    anc_data = self._get_ancillary_data_for_acquisition(curr_acq)
    select_pt_func = getattr(multiobjective_gpb_acquisitions.asy,
                       self._get_name_in_multiobjective_acquisition_module(curr_acq))
    qinfo_hp_tune_method = self.gp_processors[0].hp_tune_method
    qinfo = Namespace(curr_acq=curr_acq, hp_tune_method=qinfo_hp_tune_method)
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    else:
      next_eval_point = select_pt_func(self.gps, anc_data)
      qinfo.point = next_eval_point
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determine the next batch of evaluation points. """
    curr_acq = self._get_next_acq()
    anc_data = self._get_ancillary_data_for_acquisition(curr_acq)
    select_pt_func = getattr(multiobjective_gpb_acquisitions.syn,
                       self._get_name_in_multiobjective_acquisition_module(curr_acq))
    qinfo_hp_tune_method = self.gp_processors[0].hp_tune_method
    if self.is_an_mf_method():
      raise NotImplementedError('Not Implemented synchronous mf yet!')
    else:
      next_batch_of_eval_points = select_pt_func(batch_size, self.gps, anc_data)
      qinfos = [Namespace(point=pt, hp_tune_method=qinfo_hp_tune_method)
                for pt in next_batch_of_eval_points]
    return qinfos

  def _main_loop_pre_boca(self):
    """ Things to be done before each iteration of the optimisation loop for BOCA. """
    raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)

  def _main_loop_pre(self):
    """ Things to be done before each iteration of the optimisation loop. """
    self._set_next_gp() # set the next GP
    # For BOCA
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
      # if self.options.mf_strategy == 'boca':
      #   self._main_loop_pre_boca()

# MultiObjectiveGPBandit class ends here =================================================


# A Multi Objective Euclidean GP Bandit
class EuclideanMultiObjectiveGPBandit(MultiObjectiveGPBandit):
  """ A GP Bandit for Euclidean Spaces. """

  # Constructor.
  def __init__(self, multi_func_caller, worker_manager, is_mf=False,
               options=None, reporter=None):
    """ Constructor. """
    if options is None:
      reporter = get_reporter(reporter)
      if is_mf:
        raise NotImplementedError("MF support for MO not implemented yet")
        # all_args = get_all_mf_gp_bandit_args_from_gp_args(euclidean_mf_gp_args)
      else:
        all_args = get_all_euc_moo_gp_bandit_args()
      options = load_options(all_args, reporter=reporter)
    super(EuclideanMultiObjectiveGPBandit, self).__init__(multi_func_caller,
      worker_manager, is_mf=is_mf, options=options, reporter=reporter)

  def _get_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the Multi-fidelity GP Fitter. Can be overridded by a child class. """
    raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)

  def _get_non_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the NOn-Multi-fidelity GP Fitter. Can be overridded by a child class. """
    # pylint: disable=no-member
    options = Namespace(**vars(self.options))
    if use_additive:
      options.use_additive_gp = use_additive
    if use_additive and options.kernel_type == 'esp':
      options.kernel_type = options.domain_esp_kernel_type
    return EuclideanGPFitter(reg_data[0], reg_data[1],
                             options=options, reporter=self.reporter)


  def _domain_specific_multi_opt_method_set_up(self):
    """ Some set up for the EuclideanGPBandit class. """
    self.add_gps = None
    # Flag for creating extra add_gp
    self.req_add_gp = False
    if self.is_an_mf_method():
      raise NotImplementedError("MF support for MO has not been implemented yet")
    else:
      if not self.options.use_additive_gp and 'add_ucb' in self.acqs_to_use:
        self.req_add_gp = True

  def _domain_specific_build_new_gps(self):
    """ Builds an additive GP if required and stores it in self.add_gp. """
    if self.req_add_gp:
      self.add_gp_processors = []
      for i in range(self.multi_func_caller.num_funcs):
        reg_data = self._get_moo_gp_reg_data(i)
        add_gp_fitter = self._get_gp_fitter(reg_data, use_additive=True)
        # Fits gp and adds it to add_gp_processor
        add_gp_fitter.fit_gp_for_gp_bandit(self.options.build_new_model_every)
        add_gp_processor = Namespace()
        add_gp_processor.gp_fitter = add_gp_fitter
        self.add_gp_processors.append(add_gp_processor)
      self.add_gps = None # Mostly to avoid bugs

  def _domain_specific_add_data_to_gp(self, new_data):
    """ Adds data from the child class to the EuclideanGP """
    if hasattr(self, 'add_gp_processors'):
      if self.is_an_mf_method():
        raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
      else:
        for add_gp_processor in self.add_gp_processors:
          if add_gp_processor.fit_type == 'fitted_gp' and self.add_gps is not None:
            split_data = _split_multi_objective_into_single_objective(
                           new_data[0], new_data[1])
            for i, add_gp in enumerate(self.add_gps):
              data_pair = split_data[i]
              add_gp.add_data_multiple(data_pair[0], data_pair[1])

  def _domain_specific_multi_opt_method_update_history(self, qinfo, num_improvements):
    """ Update history for GP bandit specific statistics. """
    if hasattr(self, 'add_gp_processors') and hasattr(qinfo, 'hp_tune_method'):
      for agp in self.add_gp_processors:
        agp.gp_fitter.update_hp_tune_method_weight(qinfo.hp_tune_method, num_improvements)

  def _domain_specific_set_next_gp(self):
    self.add_gps = []
    if hasattr(self, 'add_gp_processors'):
      for i, add_gp_processor in enumerate(self.add_gp_processors):
        ret = add_gp_processor.gp_fitter.get_next_gp()
        add_gp_processor.fit_type = ret[0]
        add_gp_processor.hp_tune_method = ret[1]
        add_gp = ret[2]
        self.add_gps.append(add_gp)
        if add_gp_processor.fit_type in ['sample_hps_with_probs',
                                         'post_sample_hps_with_probs']:
          reg_data = self._get_moo_gp_reg_data(i)
          self._child_set_add_gp_data(reg_data, i)

  def _child_set_add_gp_data(self, reg_data, obj_ind):
    """ Set data in child. Can be overridden by a child class. """
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    else:
      self.add_gps[obj_ind].set_data(reg_data[0], reg_data[1], build_posterior=True)

  def _determine_next_query(self):
    """ Determine the next point for evaluation. Override for Euclidean spaces. """
    curr_acq = self._get_next_acq()
    anc_data = self._get_ancillary_data_for_acquisition(curr_acq)
    select_pt_func = getattr(multiobjective_gpb_acquisitions.asy,
                       self._get_name_in_multiobjective_acquisition_module(curr_acq))
    if curr_acq == 'add_ucb':
      qinfo_hp_tune_method = self.add_gp_processors[0].hp_tune_method
    else:
      qinfo_hp_tune_method = self.gp_processors[0].hp_tune_method
    qinfo = Namespace(curr_acq=curr_acq, hp_tune_method=qinfo_hp_tune_method)
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    else:
      if self.add_gps is None or curr_acq != 'add_ucb':
        next_eval_point = select_pt_func(self.gps, anc_data)
      else:
        next_eval_point = select_pt_func(self.add_gps, anc_data)
      qinfo.point = next_eval_point
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determine the next batch of evaluation points. Override for Euclidean spaces. """
    curr_acq = self._get_next_acq()
    anc_data = self._get_ancillary_data_for_acquisition(curr_acq)
    select_pt_func = getattr(multiobjective_gpb_acquisitions.syn,
                       self._get_name_in_multiobjective_acquisition_module(curr_acq))
    if curr_acq == 'add_ucb':
      qinfo_hp_tune_method = self.add_gp_processors[0].hp_tune_method
    else:
      qinfo_hp_tune_method = self.gp_processors[0].hp_tune_method
    if self.is_an_mf_method():
      raise NotImplementedError('Not Implemented synchronous mf yet!')
    else:
      if self.add_gps is None or curr_acq != 'add_ucb':
        next_batch_of_eval_points = select_pt_func(batch_size, self.gps, anc_data)
      else:
        next_batch_of_eval_points = select_pt_func(batch_size, self.add_gps, anc_data)
      qinfos = [Namespace(point=pt, hp_tune_method=qinfo_hp_tune_method)
                for pt in next_batch_of_eval_points]
    return qinfos

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    if self.is_an_mf_method():
      return get_euclidean_initial_qinfos(self.options.init_method, num_init_evals, \
               self.domain.bounds, self.options.fidel_init_method, \
               self.fidel_space.bounds, self.func_caller.fidel_to_opt, \
               self.options.init_set_to_fidel_to_opt_with_prob)
    else:
      return get_euclidean_initial_qinfos(self.options.init_method, num_init_evals,
                                          self.domain.bounds)

  def _domain_specific_acq_opt_set_up(self):
    """ Set up acquisition optimisation for the child class. """
    if self.acq_opt_method.lower() in ['direct']:
      self._set_up_euc_acq_opt_direct()
    elif self.acq_opt_method.lower() in ['pdoo']:
      self._set_up_euc_acq_opt_pdoo()
    elif self.acq_opt_method.lower() == 'rand':
      self._set_up_euc_acq_opt_rand()
    else:
      raise NotImplementedError('Not implemented acquisition optimisation for %s yet.'%( \
                                self.acq_opt_method))

  # Any of these set up methods can be overridden by a child class -------------------
  def _set_up_euc_acq_opt_direct(self):
    """ Sets up optimisation for acquisition using direct/pdoo. """
    if self.get_acq_opt_max_evals is None:
      lead_const = 1 * min(5, self.domain.get_dim())**2
      self.get_acq_opt_max_evals = lambda t: np.clip(lead_const * np.sqrt(min(t, 1000)),
                                                     1000, 3e4)

  def _set_up_euc_acq_opt_pdoo(self):
    """ Sets up optimisation for acquisition using direct/pdoo. """
    if self.get_acq_opt_max_evals is None:
      lead_const = 2 * min(5, self.domain.get_dim())**2
      self.get_acq_opt_max_evals = lambda t: np.clip(lead_const * np.sqrt(min(t, 1000)),
                                                     2000, 6e4)

  def _set_up_euc_acq_opt_rand(self):
    """ Sets up optimisation for acquisition using random search. """
    if self.get_acq_opt_max_evals is None:
      lead_const = 10 * min(5, self.domain.get_dim())**2
      self.get_acq_opt_max_evals = lambda t: np.clip(lead_const * np.sqrt(min(t, 1000)),
                                                     2000, 3e4)


class CPMultiObjectiveGPBandit(MultiObjectiveGPBandit):
  """ A GP Bandit class on Cartesian product spaces. """

  def __init__(self, multi_func_caller, worker_manager, is_mf=False,
               domain_dist_computers=None, options=None, reporter=None):
    """ Constructor. """
    if options is None:
      reporter = get_reporter(reporter)
      if is_mf:
        raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
      else:
        all_args = get_all_cp_moo_gp_bandit_args()
      options = load_options(all_args, reporter)
    self.domain_dist_computers = domain_dist_computers
    super(CPMultiObjectiveGPBandit, self).__init__(multi_func_caller, worker_manager,
                                          is_mf=is_mf, options=options, reporter=reporter)

  def _domain_specific_multi_opt_method_set_up(self):
    """ Any domain specific set up for multi-objective GP bandit. """
    self.domain_lists_of_dists = None
    if self.domain_dist_computers is None:
      self.domain_dist_computers = [None] * self.domain.num_domains
    self.kernel_params_for_each_domain = [{} for _ in range(self.domain.num_domains)]
    # Create a Dummy GP Fitter so that we can get the mislabel and struct coeffs for
    # otmann.
    if self.is_an_mf_method():
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    else:
      dummy_gp_fitter = CPGPFitter([], [], self.func_caller.domain,
         domain_kernel_ordering=self.func_caller.domain_orderings.kernel_ordering,
         domain_lists_of_dists=None,
         domain_dist_computers=None,
         options=self.options, reporter=self.reporter)
    # Pre-compute distances for all sub-domains in domain - not doing for fidel_space
    # since we don't expect pre-computing distances will be necessary there.
    for idx, dom in enumerate(self.domain.list_of_domains):
      if dom.get_type() == 'neural_network' and self.domain_dist_computers[idx] is None:
        from nn.otmann import get_otmann_distance_computer_from_args
        otm_mislabel_coeffs =  \
          dummy_gp_fitter.domain_kernel_params_for_each_domain[idx].otmann_mislabel_coeffs
        otm_struct_coeffs =  \
          dummy_gp_fitter.domain_kernel_params_for_each_domain[idx].otmann_struct_coeffs
        self.domain_dist_computers[idx] = get_otmann_distance_computer_from_args(
          dom.nn_type, self.options.otmann_non_assignment_penalty,
          otm_mislabel_coeffs, otm_struct_coeffs, self.options.otmann_dist_type)
        self.kernel_params_for_each_domain[idx]['otmann_dist_type'] = \
          self.options.otmann_dist_type

  def _get_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the Multi-fidelity GP Fitter. Can be overridded by a child class. """
    raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)

  def _get_non_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the NOn-Multi-fidelity GP Fitter. Can be overridded by a child class. """
    return CPGPFitter(reg_data[0], reg_data[1], self.func_caller.domain,
             domain_kernel_ordering=self.func_caller.domain_orderings.kernel_ordering,
             domain_lists_of_dists=self.domain_lists_of_dists,
             domain_dist_computers=self.domain_dist_computers,
             options=self.options, reporter=self.reporter)

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    if self.is_an_mf_method():
      return get_cp_domain_initial_qinfos(self.domain, num_init_evals,
        fidel_space=self.fidel_space, fidel_to_opt=self.func_caller.fidel_to_opt,
        set_to_fidel_to_opt_with_prob=self.options.init_set_to_fidel_to_opt_with_prob,
        dom_euclidean_sample_type='latin_hc',
        dom_integral_sample_type='latin_hc',
        dom_nn_sample_type='rand',
        fidel_space_euclidean_sample_type='latin_hc',
        fidel_space_integral_sample_type='latin_hc',
        fidel_space_nn_sample_type='rand')
    else:
      return get_cp_domain_initial_qinfos(self.domain, num_init_evals,
                                          dom_euclidean_sample_type='latin_hc',
                                          dom_integral_sample_type='latin_hc',
                                          dom_nn_sample_type='rand')

  # Set up for acquisition optimisation --------------------------------------------
  def _domain_specific_acq_opt_set_up(self):
    """ Set up acquisition optimisation for the child class. """
    if self.acq_opt_method.lower() in ['direct']:
      self._set_up_cp_acq_opt_direct()
    elif self.acq_opt_method.lower() in ['pdoo']:
      self._set_up_cp_acq_opt_pdoo()
    elif self.acq_opt_method.lower() == 'rand':
      self._set_up_cp_acq_opt_rand()
    elif self.acq_opt_method.lower().startswith('ga'):
      self._set_up_cp_acq_opt_ga()
    else:
      raise ValueError('Unrecognised acq_opt_method "%s".'%(self.acq_opt_method))

  # Any of these set up methods can be overridden by a child class -------------------
  def _set_up_cp_acq_opt_with_params(self, lead_const, min_iters, max_iters):
    """ Set up acquisition optimisation with params. """
    if self.get_acq_opt_max_evals is None:
      dim_factor = lead_const * min(5, self.domain.get_dim())**2
      self.get_acq_opt_max_evals = lambda t: np.clip(dim_factor * np.sqrt(min(t, 1000)),
                                                     min_iters, max_iters)

  def _set_up_cp_acq_opt_direct(self):
    """ Sets up optimisation for acquisition using direct/pdoo. """
    self._set_up_cp_acq_opt_with_params(2, 1000, 3e4)

  def _set_up_cp_acq_opt_pdoo(self):
    """ Sets up optimisation for acquisition using direct/pdoo. """
    self._set_up_cp_acq_opt_with_params(2, 2000, 6e4)

  def _set_up_cp_acq_opt_rand(self):
    """ Set up optimisation for acquisition using rand. """
    self._set_up_cp_acq_opt_with_params(1, 1000, 3e4)

  def _set_up_cp_acq_opt_ga(self):
    """ Set up optimisation for acquisition using rand. """
    domain_types = [dom.get_type() for dom in self.domain.list_of_domains]
    if 'neural_network' in domain_types:
      # Because Neural networks can be quite expensive
      self._set_up_cp_acq_opt_with_params(1, 500, 2e4)
    else:
      self._set_up_cp_acq_opt_with_params(1, 1000, 3e4)





# Utilities
def _split_multi_objective_into_single_objective(X, Y):
  """ Splits k-objective observations into k sets of
      single objective observations.
  """
  num_obj = len(Y[0])
  for y in Y:
    assert len(y) == num_obj
  data_pairs = []
  for i in range(num_obj):
    pair = (X, [y[i] for y in Y])
    data_pairs.append(pair)
  return data_pairs


# APIs for Euclidean GP Bandit optimisation. ---------------------------------------------
# 1. Optimisation from a FunctionCaller object.
def multiobjective_gpb_from_multi_func_caller(multi_func_caller, worker_manager,
  max_capital, is_mf, mode=None, acq=None, moo_strategy='moors',
  moo_strategy_parameters=None, mf_strategy=None, domain_add_max_group_size=-1,
  options=None, reporter='default'):
  # pylint: disable=too-many-arguments
  """ Multi-objective GPBandit from a function caller.
  """
  # pylint: disable=too-many-branches
  reporter = get_reporter(reporter)
  if is_mf:
    raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
  else:
    if isinstance(multi_func_caller.domain, domains.EuclideanDomain):
      optimiser_constructor = EuclideanMultiObjectiveGPBandit
      dflt_list_of_options = get_all_euc_moo_gp_bandit_args()
    elif isinstance(multi_func_caller.domain, domains.CartesianProductDomain):
      optimiser_constructor = CPMultiObjectiveGPBandit
      dflt_list_of_options = get_all_cp_moo_gp_bandit_args()
    else:
      raise ValueError('GPBandit not implemented for domain of type %s.'%( \
                       type(multi_func_caller.domain)))
  # Load options
  if options is None:
    options = load_options(dflt_list_of_options, reporter=reporter)
  if acq is not None:
    options.acq = acq
  if mode is not None:
    options.mode = mode
  if mf_strategy is not None:
    options.mf_strategy = mf_strategy
  if moo_strategy == 'moors':
    if moo_strategy_parameters is not None:
      if hasattr(moo_strategy_parameters, 'moors_weight_sampler') and \
        moo_strategy_parameters.moors_weight_sampler is not None:
        options.moors_weight_sampler = moo_strategy_parameters.moors_weight_sampler
      if hasattr(moo_strategy_parameters, 'moors_reference_point') and \
        moo_strategy_parameters.moors_reference_point is not None:
        options.moors_reference_point = moo_strategy_parameters.moors_reference_point
  else:
    raise ValueError('Unknown MOO Strategy %s.'%(moo_strategy))
  # Domain Specific Parameters
  if isinstance(multi_func_caller.domain, domains.EuclideanDomain) \
    and domain_add_max_group_size >= 0:
    if is_mf:
      raise NotImplementedError(_NO_MF_FOR_MOGPB_ERR_MSG)
    else:
      options.use_additive_gp = True
      if domain_add_max_group_size > 0:
        options.add_max_group_size = domain_add_max_group_size
  # create optimiser and return
  optimiser = optimiser_constructor(multi_func_caller, worker_manager, is_mf=is_mf,
                                    options=options, reporter=reporter)
  return optimiser.optimise(max_capital)


def cp_multiobjective_gpb_from_raw_args(raw_funcs, domain_config_file, *args, **kwargs):
  """ A random multi-objective optimiser on CP spaces. """
  # pylint: disable=no-member
  cp_dom, orderings = load_cp_domain_from_config_file(domain_config_file)
  proc_funcs = [get_processed_func_from_raw_func_for_cp_domain(
                  rf, cp_dom, orderings.index_ordering, orderings.dim_ordering)
                for rf in raw_funcs]
  multi_func_caller = CPMultiFunctionCaller(proc_funcs, cp_dom, raw_funcs=raw_funcs,
                                            domain_orderings=orderings)
  return multiobjective_gpb_from_multi_func_caller(multi_func_caller, *args, **kwargs)




"""
  Utilities to modify a given neural network and obtain a new one.
  --kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=star-args
# pylint: disable=too-many-branches

from argparse import Namespace
from copy import deepcopy
import numpy as np
# Local imports
from .neural_network import ConvNeuralNetwork, MultiLayerPerceptron, MLP_RECTIFIERS, \
       MLP_SIGMOIDS, is_a_pooling_layer_label, is_a_conv_layer_label,\
       CNNImageSizeMismatchException, CNNNoConvAfterIPException
from ..utils.general_utils import reorder_list_or_array, reorder_rows_and_cols_in_matrix
from ..utils.option_handler import get_option_specs, load_options
from ..utils.reporters import get_reporter

_DFLT_CHANGE_FRAC = 0.125
_DFLT_CHANGE_NUM_UNITS_SPAWN = 'all'
_DFLT_CHANGE_LAYERS_SPAWN = 20
_DFLT_NUM_SINGLE_STEP_MODIFICATIONS = 'all'
_DFLT_NUM_TWO_STEP_MODIFICATIONS = 0
_DFLT_NUM_THREE_STEP_MODIFICATIONS = 0
_DFLT_WEDGE_LAYER_CNN_CANDIDATES = ['conv3', 'conv5', 'conv7', 'res3', 'res5', 'res7']
_DFLT_WEDGE_LAYER_MLP_CANDIDATES = MLP_RECTIFIERS + MLP_SIGMOIDS
_DFLT_SIGMOID_SWAP = MLP_SIGMOIDS
_DFLT_RECTIFIER_SWAP = MLP_RECTIFIERS

_PRIMITIVE_PROB_MASSES = {'inc_single': 0.1,
                          'dec_single': 0.1,
                          'inc_en_masse': 0.1,
                          'dec_en_masse': 0.1,
                          'swap_layer': 0.2,
                          'wedge_layer': 0.1,
                          'remove_layer': 0.1,
                          'branch': 0.2,
                          'skip': 0.2,
                         }

nn_modifier_args = [
  # Change fractions for increasing the number of units in layers.
  get_option_specs('single_inc_change_frac', False, _DFLT_CHANGE_FRAC,
    'Default change fraction when increasing a single layer.'),
  get_option_specs('single_dec_change_frac', False, _DFLT_CHANGE_FRAC,
    'Default change fraction when decreasing a single layer.'),
  get_option_specs('en_masse_inc_change_frac', False, _DFLT_CHANGE_FRAC,
    'Default change fraction when increasing layers en_masse.'),
  get_option_specs('en_masse_dec_change_frac', False, _DFLT_CHANGE_FRAC,
    'Default change fraction when decreasing layers en_masse.'),
  # Number of networks to spawn by changing number of units in a single layer.
  get_option_specs('spawn_single_inc_num_units', False, _DFLT_CHANGE_NUM_UNITS_SPAWN,
    'Default number of networks to spawn by increasing # units in a single layer.'),
  get_option_specs('spawn_single_dec_num_units', False, _DFLT_CHANGE_NUM_UNITS_SPAWN,
    'Default number of networks to spawn by decreasing # units in a single layer.'),
  # Number of networks to spawn by adding or deleting a single layer.
  get_option_specs('spawn_add_layer', False, _DFLT_CHANGE_LAYERS_SPAWN,
    'Default number of networks to spawn by adding a layer.'),
  get_option_specs('spawn_del_layer', False, _DFLT_CHANGE_LAYERS_SPAWN,
    'Default number of networks to spawn by deleting a layer.'),
  # Number of double/triple step candidates - i.e. applications of basic primitives
  # twice/thrice before executing candidates
  get_option_specs('num_single_step_modifications', False,
    _DFLT_NUM_SINGLE_STEP_MODIFICATIONS,
    'Default number of networks to spawn via single step primitives.'),
  get_option_specs('num_two_step_modifications', False,
    _DFLT_NUM_TWO_STEP_MODIFICATIONS,
    'Default number of networks to spawn via two step primitives.'),
  get_option_specs('num_three_step_modifications', False,
    _DFLT_NUM_THREE_STEP_MODIFICATIONS,
    'Default number of networks to spawn via three step primitives.'),
  ]


# Generic utilities we will need in all functions below ==================================
def get_copies_from_old_nn(nn):
  """ Gets copies of critical parameters of the old network. """
  layer_labels = deepcopy(nn.layer_labels)
  num_units_in_each_layer = deepcopy(nn.num_units_in_each_layer)
  conn_mat = deepcopy(nn.conn_mat)
  mandatory_child_attributes = Namespace()
  for mca_str in nn.mandatory_child_attributes:
    mca_val = deepcopy(getattr(nn, mca_str))
    setattr(mandatory_child_attributes, mca_str, mca_val)
  return layer_labels, num_units_in_each_layer, conn_mat, mandatory_child_attributes

def get_new_nn(old_nn, layer_labels, num_units_in_each_layer, conn_mat,
               mandatory_child_attributes):
  """ Returns a new neural network of the same type as old_nn. """
  known_nn_class = True
  try:
    if old_nn.nn_class == 'cnn':
      new_cnn = ConvNeuralNetwork(layer_labels, conn_mat, num_units_in_each_layer,
                                  mandatory_child_attributes.strides,
                                  old_nn.all_layer_label_classes,
                                  old_nn.layer_label_similarities)
      return new_cnn
    elif old_nn.nn_class.startswith('mlp'):
      return MultiLayerPerceptron(old_nn.nn_class[4:], layer_labels, conn_mat,
                                  num_units_in_each_layer, old_nn.all_layer_label_classes,
                                  old_nn.layer_label_similarities)
    else:
      known_nn_class = False
  except (CNNImageSizeMismatchException, CNNNoConvAfterIPException, AssertionError):
    return None
  if not known_nn_class:
    raise ValueError('Unidentified nn_class %s.'%(old_nn.nn_class))

def add_layers_to_end_of_conn_mat(conn_mat, num_add_layers):
  """ Adds layers with no edges and returns. """
  new_num_layers = conn_mat.shape[0] + num_add_layers
  conn_mat.resize((new_num_layers, new_num_layers))
  return conn_mat


# Change architecture of the network
# ========================================================================================

# Add a layer ----------------------------------------------------------------------------
def wedge_layer(nn, layer_type, units_in_layer, layer_before, layer_after,
                new_layer_attributes=None):
  """ Wedges a layer of type layer_type after the layer given in layer_before. The
      output of the layer in layer_before goes to the new layer and the output of the
      new layer goes to layer_after. If an edge existed between layer_before and
      layer_after, it is removed. """
  layer_labels, num_units_in_each_layer, conn_mat, mandatory_child_attributes = \
    get_copies_from_old_nn(nn)
  layer_labels.append(layer_type)
  num_units_in_each_layer = np.append(num_units_in_each_layer, units_in_layer)
  if nn.nn_class == 'cnn':
    mandatory_child_attributes.strides.append(new_layer_attributes.stride)
  conn_mat = add_layers_to_end_of_conn_mat(conn_mat, 1)
  conn_mat[layer_before, -1] = 1
  conn_mat[-1, layer_after] = 1
  conn_mat[layer_before, layer_after] = 0
  return get_new_nn(nn, layer_labels, num_units_in_each_layer, conn_mat,
                    mandatory_child_attributes)

def _get_non_None_elements(iter_of_vals):
  """ Returns non None values. """
  return [x for x in iter_of_vals if x is not None]

def _determine_num_units_for_wedge_layer(nn, edge):
  """ Determines the number of layers for wedging a layer. This is usually the average
      of the parent (edge[0]) and child (edge[1]).
  """
  edge_num_layers = _get_non_None_elements(
                      [nn.num_units_in_each_layer[idx] for idx in edge])
  if len(edge_num_layers) > 0:
    return round(sum(edge_num_layers) / len(edge_num_layers))
  else:
    parents = nn.get_parents(edge[0])
    if len(parents) == 0:
      # Means you have reached the input node
      ip_children = nn.get_children(edge[0])
      children_num_units = _get_non_None_elements(
                             [nn.num_units_in_each_layer[ch] for ch in ip_children])
      if len(children_num_units) == 0:
        # Create a layer with 16 units
        children_num_units = [16]
      return sum(children_num_units) / len(children_num_units)
    else:
      parent_num_units = _get_non_None_elements(
                          [nn.num_units_in_each_layer[par] for par in parents])
      if len(parent_num_units) > 0:
        return sum(parent_num_units) / len(parent_num_units)
      else:
        par_num_units = []
        for par in parents:
          par_num_units.append(_determine_num_units_for_wedge_layer(nn, (par, edge[0])))
        par_num_units = _get_non_None_elements(par_num_units)
        return sum(par_num_units) / len(par_num_units)

def get_list_of_wedge_layer_modifiers(nn, num_modifications='all',
                                      internal_layer_type_candidates=None,
                                      choose_pool_with_prob=0.05,
                                      choose_stride_2_with_prob=0.05):
  """ Returns a list of operations for adding a layer in between two layers. """
  # A local function for creating a modifier
  def _get_wedge_modifier(_layer_type, _num_units, _edge, _nl_attributes):
    """ Returns a modifier which wedges an edge between the edge. """
    return lambda arg_nn: wedge_layer(arg_nn, _layer_type, _num_units,
                                       _edge[0], _edge[1], _nl_attributes)
  # Pre-process arguments
  nn_is_a_cnn = nn.nn_class == 'cnn'
  if internal_layer_type_candidates is None:
    if nn_is_a_cnn:
      internal_layer_type_candidates = _DFLT_WEDGE_LAYER_CNN_CANDIDATES
    else:
      internal_layer_type_candidates = _DFLT_WEDGE_LAYER_MLP_CANDIDATES
  if not nn_is_a_cnn:
    choose_pool_with_prob = 0
  all_edges = nn.get_edges()
  num_modifications = len(all_edges) if num_modifications == 'all' else num_modifications
  op_layer_idx = nn.get_op_layer_idx() # Output layer
  ip_layer_idx = nn.get_ip_layer_idx() # Input layer
  # We won't change this below so keep it as it is
  nonconv_nl_attrs = Namespace(stride=None)
  conv_nl_attrs_w_stride_1 = Namespace(stride=1)
  conv_nl_attrs_w_stride_2 = Namespace(stride=2)
  # Iterate through all edges
  ret = []
  for edge in all_edges:
    curr_layer_type = None
    # First handle the edges cases
    if edge[1] == op_layer_idx:
      continue
    elif nn_is_a_cnn and nn.layer_labels[edge[0]] == 'fc':
      curr_layer_type = 'fc'
      curr_num_units = nn.num_units_in_each_layer[edge[0]]
      nl_attrs = nonconv_nl_attrs
    elif not nn_is_a_cnn and edge[1] == op_layer_idx:
      # Don't add new layers just before the output for MLPs
      continue
    elif edge[0] == ip_layer_idx and nn_is_a_cnn:
      curr_pool_prob = 0 # No pooling layer right after the input for a CNN
    else:
      curr_pool_prob = choose_pool_with_prob

    if curr_layer_type is None:
      if np.random.random() < curr_pool_prob:
        curr_layer_candidates = ['avg-pool', 'max-pool']
      else:
        curr_layer_candidates = internal_layer_type_candidates
      curr_layer_type = np.random.choice(curr_layer_candidates, 1)[0]
      if curr_layer_type in ['max-pool', 'avg-pool', 'linear', 'softmax']:
        curr_num_units = None
      else:
        curr_num_units = _determine_num_units_for_wedge_layer(nn, edge)
      # Determine stride
      if is_a_conv_layer_label(curr_layer_type):
        nl_attrs = conv_nl_attrs_w_stride_2 if \
          np.random.random() < choose_stride_2_with_prob else conv_nl_attrs_w_stride_1
      else:
        nl_attrs = nonconv_nl_attrs
    ret.append(_get_wedge_modifier(curr_layer_type, curr_num_units, edge, nl_attrs))
    # Break if more than the number of modifications
    if len(ret) >= num_modifications:
      break
  return ret


# Removing a layer -----------------------------------------------------------------------
def remove_layer(nn, del_idx, additional_edges, new_strides=None):
  """ Deletes the layer indicated in del_idx and adds additional_edges specified
      in additional_edges. """
  layer_labels, num_units_in_each_layer, conn_mat, mandatory_child_attributes = \
    get_copies_from_old_nn(nn)
  # First add new edges to conn_mat and remove edges to and from del_idx
  for add_edge in additional_edges:
    conn_mat[add_edge[0], add_edge[1]] = 1
  conn_mat[del_idx, :] = 0
  conn_mat[:, del_idx] = 0
  # Now reorder everything so that del_idx is at the end
  all_idxs = list(range(len(layer_labels)))
  new_order = all_idxs[:del_idx] + all_idxs[del_idx+1:] + [del_idx]
  # Now reorder everything so that the layer to be remove is at the end
  layer_labels = reorder_list_or_array(layer_labels, new_order)
  num_units_in_each_layer = reorder_list_or_array(num_units_in_each_layer, new_order)
  conn_mat = reorder_rows_and_cols_in_matrix(conn_mat, new_order)
  # remove layer
  layer_labels = layer_labels[:-1]
  num_units_in_each_layer = num_units_in_each_layer[:-1]
  conn_mat = conn_mat[:-1, :-1]
  # Strides for a convolutional network
  if nn.nn_class == 'cnn':
    new_strides = new_strides if new_strides is not None else \
                  mandatory_child_attributes.strides
    mandatory_child_attributes.strides = reorder_list_or_array(
      new_strides, new_order)
    mandatory_child_attributes.strides = mandatory_child_attributes.strides[:-1]
  return get_new_nn(nn, layer_labels, num_units_in_each_layer, conn_mat,
                    mandatory_child_attributes)


def get_list_of_remove_layer_modifiers(old_nn):
  """ Returns a list of primitives which remove a layer from a neural network. """
  # pylint: disable=too-many-locals
  # A local function to return the modifier
  if old_nn.num_processing_layers == 0:
    # Don't delete any layers if there are no processing layers.
    return []
  def _get_remove_modifier(_del_idx, _add_edges, *_args, **_kwargs):
    """ Returns a modifier which deletes _del_idx and adds _add_edges. """
    return lambda arg_nn: remove_layer(arg_nn, _del_idx, _add_edges, *_args, **_kwargs)
  # Now check every layer
  ret = []
  for idx, ll in enumerate(old_nn.layer_labels):
    if ll in ['ip', 'op', 'softmax', 'linear']: # Don't delete any of these layers
      continue
    curr_parents = old_nn.get_parents(idx)
    parent_labels = [old_nn.layer_labels[par_idx] for par_idx in curr_parents]
    if ll == 'fc' and (not parent_labels == ['fc'] * len(parent_labels)):
      # If the parents of a fc layer are also not fc then do not delete
      continue
    curr_children = old_nn.get_children(idx)
    if old_nn.nn_class == 'cnn' and \
       old_nn.pre_img_inv_sizes[idx] != old_nn.post_img_inv_sizes[idx]:
      change_stride_idxs = None
      # First check if the children are modifiable
      child_strides = [old_nn.strides[ch_idx] for ch_idx in curr_children]
      if child_strides == [1] * len(curr_children):
        change_stride_idxs = curr_children # we will change the strides of the children
      if change_stride_idxs is None:
        parent_strides = [old_nn.strides[par_idx] for par_idx in curr_parents]
        if parent_strides == [1] * len(curr_parents):
          change_stride_idxs = curr_parents
      # If we have successfuly identified children/parents which we can modify, great!
      # Otherwise, lets not change anything and hope that it
      # does not break anything. If it does, there is an exception to handle this.
      if change_stride_idxs is not None:
        new_strides = deepcopy(old_nn.strides)
        for csi in change_stride_idxs:
          new_strides[csi] = 2
      else:
        new_strides = None
    else:
      new_strides = None
    # Now delete the layer and add new adges
    num_children_on_each_parent = [len(old_nn.get_children(par_idx)) for par_idx in
                                   curr_parents]
    num_parents_on_each_child = [len(old_nn.get_parents(ch_idx)) for ch_idx in
                                 curr_children]
    must_add_children = [curr_children[i] for i in range(len(curr_children)) if
                         num_parents_on_each_child[i] == 1]
    must_add_parents = [curr_parents[i] for i in range(len(curr_parents)) if
                        num_children_on_each_parent[i] == 1]
    num_must_add_children = len(must_add_children)
    num_must_add_parents = len(must_add_parents)
    np.random.shuffle(must_add_children)
    np.random.shuffle(must_add_parents)
    add_edges = []
    for _ in range(min(num_must_add_children, num_must_add_parents)):
      add_edges.append((must_add_parents.pop(), must_add_children.pop()))
    # Add edges for left over children/parents
    if num_must_add_children > num_must_add_parents:
      diff = num_must_add_children - num_must_add_parents
      cand_parents = list(np.random.choice(curr_parents, diff))
      for _ in range(diff):
        add_edges.append((cand_parents.pop(), must_add_children.pop()))
    if num_must_add_parents > num_must_add_children:
      diff = num_must_add_parents - num_must_add_children
      cand_children = list(np.random.choice(curr_children, diff))
      for _ in range(diff):
        add_edges.append((must_add_parents.pop(), cand_children.pop()))
    ret.append(_get_remove_modifier(idx, add_edges, new_strides))
  return ret


# Branching modifications ----------------------------------------------------------------
def create_duplicate_branch(nn, path, keep_layer_with_prob=0.5):
  """ Creates a new network which creates a new branch between path[0] and path[-1] and
      copies all layers between. It keeps a layer in between with probability 0.5. If
      in CNNs, the layer shrinks the size of the image, then we keep it with prob 1.
  """
  layer_labels, num_units_in_each_layer, conn_mat, mandatory_child_attributes = \
    get_copies_from_old_nn(nn)
  # First decide which nodes in the path to keep
  branched_path = [path[0]]
  fc_encountered = False
  for idx in path[1: -1]:
    if idx == path[1] and nn.get_ip_layer_idx() == path[0]:
      branched_path.append(idx) # Append if the branch starts at ip and this is a child.
    elif idx == path[-2] and len(branched_path) == 1:
      branched_path.append(idx) # If this is the last layer and we have not appended yet.
    elif is_a_pooling_layer_label(nn.layer_labels[idx]) or \
      nn.layer_labels[idx] in ['linear', 'softmax'] or \
      (hasattr(nn, 'strides') and nn.strides[idx] == 2):
      branched_path.append(idx)
    elif nn.layer_labels[idx] == 'fc' and not fc_encountered:
      branched_path.append(idx)
      fc_encountered = True
    elif np.random.random() < keep_layer_with_prob:
      branched_path.append(idx)
  branched_path.append(path[-1])
  # Now create additional nodes
  num_new_nodes = len(branched_path) - 2
  layer_labels.extend([nn.layer_labels[idx] for idx in branched_path[1:-1]])
  num_units_in_each_layer = np.concatenate((num_units_in_each_layer,
    [nn.num_units_in_each_layer[idx] for idx in branched_path[1:-1]]))
  # Add edges
  new_idxs = list(range(nn.num_layers, nn.num_layers + num_new_nodes))
  conn_mat = add_layers_to_end_of_conn_mat(conn_mat, num_new_nodes)
  if num_new_nodes == 0:
    conn_mat[branched_path[0], branched_path[1]] = 1
  else:
    conn_mat[branched_path[0], new_idxs[0]] = 1
    conn_mat[new_idxs[-1], branched_path[-1]] = 1
    for new_idx in new_idxs[:-1]:
      conn_mat[new_idx, new_idx + 1] = 1
  # Add strides
  if nn.nn_class == 'cnn':
    mandatory_child_attributes.strides.extend([nn.strides[idx] for idx in
                                               branched_path[1:-1]])
  return get_new_nn(nn, layer_labels, num_units_in_each_layer, conn_mat,
                    mandatory_child_attributes)


def _get_path_for_branching_from_start_layer(nn, start_layer, min_path_length=4,
                                             end_path_prob=0.20):
  """ Returns a path which starts at start layer. """
  path = [start_layer]
  while True:
    curr_layer = path[-1]
    curr_children = nn.get_children(curr_layer)
    next_layer = int(np.random.choice(curr_children, 1))
    path.append(next_layer)
    if nn.layer_labels[next_layer] == 'op':
      break
    elif len(path) < min_path_length:
      pass
    elif np.random.random() < end_path_prob:
      break
  return path


def _get_start_layer_probs_for_branching_and_skipping(nn):
  """ Returns probabilities for the start layer to be used in branching and
      skipping primitives.
  """
  dists_from_ip = nn.get_distances_from_ip()
  start_layer_prob = []
  for layer_idx, layer_label in enumerate(nn.layer_labels):
    # We pick the first layer with distance inversely proportional to its distance from ip
    curr_layer_prob = 0 if layer_label in ['op', 'softmax', 'fc', 'linear'] else \
                      1.0 / np.sqrt(1 + dists_from_ip[layer_idx])
    start_layer_prob.append(curr_layer_prob)
  start_layer_prob = np.array(start_layer_prob)
  return start_layer_prob


def get_list_of_branching_modifiers(nn, num_modifiers=None, **kwargs):
  """ Returns a list of operators for Neural networks that create branches in the
      architecture.
  """
  if nn.num_processing_layers == 0:
    # Don't create any branches if there are no processing layers.
    return []
  # Define a local function to return the modifier
  def _get_branching_modifier(_path, *_args, **_kwargs):
    """ Returns a modifier which duplicates the path along _path. """
    return lambda arg_nn: create_duplicate_branch(arg_nn, _path, *_args, **_kwargs)
  # Some preprocessing
  num_modifiers = num_modifiers if num_modifiers is not None else 2 * nn.num_layers
  start_layer_prob = _get_start_layer_probs_for_branching_and_skipping(nn)
  ret = []
  if sum(start_layer_prob) <= 0.0:
    return ret # return immediately with an empty list
  while len(ret) < num_modifiers:
    start_layer_prob = start_layer_prob / sum(start_layer_prob)
    start_layer = int(np.random.choice(nn.num_layers, 1, p=start_layer_prob))
    path = _get_path_for_branching_from_start_layer(nn, start_layer)
    start_layer_prob[start_layer] *= 0.9 # shrink probability of picking this layer again.
    ret.append(_get_branching_modifier(path, **kwargs))
  return ret


# Skipping modifications -----------------------------------------------------------------
def create_skipped_network(nn, start_layer, end_layer, pool_layer_type='avg'):
  """ Creates a new layer with a skip connection from start_layer to end_layer.
      In a CNN, if the image sizes do not match, this creates additional pooling layers
      (either avg-pool or max-pool) to make them match.
  """
  layer_labels, num_units_in_each_layer, conn_mat, mandatory_child_attributes = \
    get_copies_from_old_nn(nn)
  if nn.nn_class != 'cnn' or \
    nn.post_img_inv_sizes[start_layer] == nn.pre_img_inv_sizes[end_layer]:
    conn_mat[start_layer, end_layer] = 1
  else:
    # Determine number of new layers, the number of units and strides in each layer.
    num_new_pool_layers = int(np.log2(nn.pre_img_inv_sizes[end_layer] /
                                      nn.post_img_inv_sizes[start_layer]))
    new_layer_idxs = list(range(nn.num_layers, nn.num_layers + num_new_pool_layers))
    num_units_in_each_layer = np.concatenate((num_units_in_each_layer,
      [None] * num_new_pool_layers))
    mandatory_child_attributes.strides.extend([None] * num_new_pool_layers)
    # Determine layer labels
    if pool_layer_type.lower().startswith('avg'):
      new_layer_type = 'avg-pool'
    elif pool_layer_type.lower().startswith('max'):
      new_layer_type = 'max-pool'
    else:
      raise ValueError('pool_layer_type should be \'avg\' or \'max\'.')
    new_layer_labels = [new_layer_type for _ in range(num_new_pool_layers)]
    layer_labels.extend(new_layer_labels)
    conn_mat = add_layers_to_end_of_conn_mat(conn_mat, num_new_pool_layers)
    # Finally, the conn_mat
    conn_mat[start_layer, new_layer_idxs[0]] = 1
    conn_mat[new_layer_idxs[-1], end_layer] = 1
    for new_idx in new_layer_idxs[:-1]:
      conn_mat[new_idx, new_idx + 1] = 1
  return get_new_nn(nn, layer_labels, num_units_in_each_layer, conn_mat,
                    mandatory_child_attributes)


def _get_end_layer_probs_for_skipping(nn, start_layer):
  """ Returns the end layer probabilities to be used in skipping. """
  dists_from_ip = nn.get_distances_from_ip()
  dists_to_op = nn.get_distances_to_op()
  is_a_cnn = nn.nn_class.startswith('cnn')
  end_layer_prob = []
  for layer_idx, layer_label in enumerate(nn.layer_labels):
    curr_layer_prob = 'assign'
    if dists_from_ip[layer_idx] - 1 <= dists_from_ip[start_layer] or \
       dists_to_op[layer_idx] + 1 >= dists_to_op[start_layer] or \
       layer_label in ['ip', 'op', 'softmax']:
      curr_layer_prob = 'no-assign'
    elif is_a_cnn and \
         nn.post_img_inv_sizes[start_layer] > nn.pre_img_inv_sizes[layer_idx]:
      # If the layer has an input image size *larger* than the output of the
      # start layer, then do not assign.
      curr_layer_prob = 'no-assign'
    elif layer_label == 'fc':
      # If its a fully connected layer, connect with this only if it is the first
      # fc layer.
      curr_layer_parent_labels = [nn.layer_labels[x] for x in nn.get_parents(layer_idx)]
      if not all([(is_a_pooling_layer_label(clpl) or is_a_conv_layer_label(clpl)) for
                  clpl in curr_layer_parent_labels]):
        curr_layer_prob = 'no-assign'
    curr_layer_prob = 0.0 if curr_layer_prob == 'no-assign' else 1.0
    end_layer_prob.append(curr_layer_prob)
  if sum(end_layer_prob) == 0:
    return None
  else:
    end_layer_prob = np.array(end_layer_prob)
    end_layer_prob = end_layer_prob / end_layer_prob.sum()
    return end_layer_prob
  return


def get_list_of_skipping_modifiers(nn, num_modifiers=None, **kwargs):
  """ Returns a list of operators for Neural networks that create branches in the
      architecture.
  """
  # Define a local function to return the modifier
  def _get_skipping_modifier(_start_layer, _end_layer, **_kwargs):
    """ Returns a modifier which adds a skip connected from start_layer to end_layer. """
    return lambda arg_nn: create_skipped_network(arg_nn, _start_layer, _end_layer,
                                                 **_kwargs)
  # Some preprocessing
  num_modifiers = num_modifiers if num_modifiers is not None else nn.num_layers
  max_num_tries = 2 * num_modifiers
  start_layer_prob = _get_start_layer_probs_for_branching_and_skipping(nn)
  start_layer_prob[0] = 0.0
  ret = []
  if sum(start_layer_prob) <= 0.0:
    return ret # return immediately with an empty list
  for _ in range(max_num_tries):
    start_layer_prob = start_layer_prob / sum(start_layer_prob)
    start_layer = int(np.random.choice(nn.num_layers, 1, p=start_layer_prob))
    end_layer_prob = _get_end_layer_probs_for_skipping(nn, start_layer)
    if end_layer_prob is None:
      continue
    end_layer = int(np.random.choice(nn.num_layers, 1, p=end_layer_prob))
    ret.append(_get_skipping_modifier(start_layer, end_layer, **kwargs))
    if len(ret) >= num_modifiers:
      break
  return ret


# Swapping -------------------------------------------------------------------------------
def swap_layer_type(nn, layer_idx, replace_with, new_stride):
  """ Swaps out the type of layer in layer_idx with replace_with """
  if nn.layer_labels[layer_idx] == replace_with:
    raise ValueError('Cannot replace layer %d with the same layer type (%s).'%(
                     layer_idx, replace_with))
  layer_labels, num_units_in_each_layer, conn_mat, mandatory_child_attributes = \
    get_copies_from_old_nn(nn)
  layer_labels[layer_idx] = replace_with
  if nn.nn_class == 'cnn':
    if is_a_pooling_layer_label(replace_with):
      num_units_in_each_layer[layer_idx] = None
    mandatory_child_attributes.strides[layer_idx] = new_stride
  return get_new_nn(nn, layer_labels, num_units_in_each_layer, conn_mat,
                    mandatory_child_attributes)

def get_list_of_swap_layer_modifiers(nn, num_modifications='all',
                                     change_stride_with_prob=0.05,
                                     rectifier_swap_candidates=None,
                                     sigmoid_swap_candidates=None):
  """ Returns a list of modifiers for swapping a layer with another. """
  # pylint: disable=too-many-statements
  # Define a local function to return the modifier
  def _get_swap_layer_modifier(_layer_idx, _replace_with, _new_stride):
    """ Returns a modifier for swapping a layer. """
    return lambda arg_nn: swap_layer_type(arg_nn, _layer_idx, _replace_with, _new_stride)
  # Preprocessing
  if nn.nn_class.startswith('mlp'):
    rectifier_swap_candidates = rectifier_swap_candidates if \
      rectifier_swap_candidates is not None else _DFLT_RECTIFIER_SWAP
    sigmoid_swap_candidates = sigmoid_swap_candidates if \
      sigmoid_swap_candidates is not None else _DFLT_SIGMOID_SWAP
  # Determine the order of the layers
  layer_order = list(range(nn.num_layers))
  if num_modifications == 'all' or num_modifications >= nn.num_layers:
    num_modifications = nn.num_layers
  else:
    np.random.shuffle(layer_order)
  # iterate through the layers and return
  ret = []
  for idx in layer_order:
    ll = nn.layer_labels[idx]
    if ll in ['ip', 'op', 'fc', 'softmax', 'linear']:
      continue # don't swap out any of these
    # Determine candidates for swapping out
    if ll == 'conv3':
      candidates = ['res3', 'res5', 'conv5', 'conv7', 'max-pool', 'avg-pool']
      cand_probs = [0.25, 0.25, 0.15, 0.25, 0.05, 0.05]
    elif ll == 'conv5':
      candidates = ['res3', 'res5', 'conv3', 'conv7', 'max-pool', 'avg-pool']
      cand_probs = [0.25, 0.25, 0.2, 0.2, 0.05, 0.05]
    elif ll == 'conv7':
      candidates = ['res3', 'res5', 'conv3', 'conv5', 'max-pool', 'avg-pool']
      cand_probs = [0.25, 0.25, 0.25, 0.15, 0.05, 0.05]
    elif ll == 'conv9':
      candidates = ['res3', 'res5', 'conv3', 'conv5', 'conv7', 'max-pool', 'avg-pool']
      cand_probs = [0.2, 0.2, 0.2, 0.2, 0.1, 0.05, 0.05]
    elif ll == 'res3':
      candidates = ['conv3', 'conv5', 'res5', 'res7', 'max-pool', 'avg-pool']
      cand_probs = [0.25, 0.25, 0.15, 0.25, 0.05, 0.05]
    elif ll == 'res5':
      candidates = ['conv3', 'conv5', 'res3', 'res7', 'max-pool', 'avg-pool']
      cand_probs = [0.25, 0.25, 0.2, 0.2, 0.05, 0.05]
    elif ll == 'res7':
      candidates = ['conv3', 'conv5', 'res3', 'res5', 'max-pool', 'avg-pool']
      cand_probs = [0.25, 0.25, 0.25, 0.15, 0.05, 0.05]
    elif ll == 'res9':
      candidates = ['conv3', 'conv5', 'res3', 'res5', 'res7', 'max-pool', 'avg-pool']
      cand_probs = [0.2, 0.2, 0.2, 0.2, 0.1, 0.05, 0.05]
    elif ll == 'avg-pool':
      candidates = ['max-pool']
      cand_probs = None
    elif ll == 'max-pool':
      candidates = ['avg-pool']
      cand_probs = None
    elif ll in MLP_RECTIFIERS:
      candidates = sigmoid_swap_candidates
      cand_probs = None
    elif ll in MLP_SIGMOIDS:
      candidates = rectifier_swap_candidates
      cand_probs = None
    else:
      raise ValueError('Unidentified layer_type: %s.'%(ll))
    # I am determining the probabilities above completely ad-hoc for reasons I don't
    # know why.
    # Choose replace_with
    if cand_probs is not None:
      cand_probs = np.array(cand_probs)
      cand_probs = cand_probs / cand_probs.sum()
    replace_with = np.random.choice(candidates, 1, p=cand_probs)[0]
    # Determine the stride
    if nn.nn_class == 'cnn':
      if is_a_pooling_layer_label(replace_with):
        new_stride = None
      elif is_a_conv_layer_label(replace_with) and is_a_pooling_layer_label(ll):
        new_stride = 2
      elif is_a_conv_layer_label(ll) and np.random.random() < change_stride_with_prob:
        new_stride = 1 if nn.strides[idx] == 2 else 2
      else:
        new_stride = nn.strides[idx]
    else:
      new_stride = None
    # Create modifier and append
    ret.append(_get_swap_layer_modifier(idx, replace_with, new_stride))
    if len(ret) >= num_modifications:
      break # Check if you have exceeded the maximum amount
  return ret

# Change number of units in a layer
# ========================================================================================
def change_num_units_in_layers(nn, change_layer_idxs, change_layer_vals):
  """ Changes the number of units in change_layer_idxs to change_layer_vals. """
  layer_labels, num_units_in_each_layer, conn_mat, mandatory_child_attributes = \
    get_copies_from_old_nn(nn)
  for i, ch_idx in enumerate(change_layer_idxs):
    if is_a_pooling_layer_label(layer_labels[ch_idx]):
      raise ValueError('Asked to change a pooling layer value. This is not allowed.')
    else:
      num_units_in_each_layer[ch_idx] = change_layer_vals[i]
  return get_new_nn(nn, layer_labels, num_units_in_each_layer, conn_mat,
                    mandatory_child_attributes)

def _get_directly_modifable_layer_idxs(nn):
  """ Returns indices that are directly modifiable in the network. """
  if isinstance(nn, ConvNeuralNetwork):
    return [i for i in range(nn.num_layers) if (nn.layer_labels[i].startswith('res') or
                                                nn.layer_labels[i].startswith('conv') or
                                                nn.layer_labels[i] == 'fc')]
  elif isinstance(nn, MultiLayerPerceptron):
    return [i for i in range(nn.num_layers) if nn.layer_labels[i] in
                                               MLP_RECTIFIERS + MLP_SIGMOIDS]
  else:
    raise ValueError('Unidentified nn type: %s'%(nn.nn_class))

def _get_change_ratio_from_change_frac(change_frac, inc_or_dec):
  """ Gets the change ratio from the change fraction, i.e. 1 +/- change_frac depending
      on inc_or_dec. """
  if inc_or_dec.lower() == 'increase':
    return 1 + abs(change_frac)
  elif inc_or_dec.lower() == 'decrease':
    return 1 - abs(change_frac)
  else:
    raise ValueError('change_ratio should be one of \'increase\' or \'decrease\'.')

def modify_several_layers(nn, inc_or_dec, layer_group_desc,
                          change_frac=_DFLT_CHANGE_FRAC):
  """ A function to increase or decrease several layers at the same time. inc_or_dec
      is a string with values 'increase' or 'decrease' and layer_group_desc is a string
      with one of the following values: '1/2', '2/2', '1/4', ... '4/4', '1/8', ... 8/8."""
  change_ratio = _get_change_ratio_from_change_frac(change_frac, inc_or_dec)
  modifiable_layers = _get_directly_modifable_layer_idxs(nn)
  num_modifiable_layers = len(modifiable_layers)
  # Now decide which groups to change
  num_groups = int(layer_group_desc[-1])
  group_idx = int(layer_group_desc[0])
  start_idx = (group_idx - 1) * num_modifiable_layers / num_groups
  end_idx = group_idx * num_modifiable_layers / num_groups
  modify_layer_idxs = modifiable_layers[start_idx:end_idx]
  modify_vals = [round(change_ratio * nn.num_units_in_each_layer[i])
                 for i in modify_layer_idxs]
  return change_num_units_in_layers(nn, modify_layer_idxs, modify_vals)

def modify_num_units_on_random_nodes(nn, num_units, inc_or_dec,
                                     change_frac=_DFLT_CHANGE_FRAC):
  """ A function to increase or decrase the number of units in a single layer. """
  change_ratio = _get_change_ratio_from_change_frac(change_frac, inc_or_dec)
  modifiable_layers = _get_directly_modifable_layer_idxs(nn)
  modify_layer_idxs = np.random.choice(modifiable_layers,
                                       max(num_units, modifiable_layers))
  modify_vals = [round(change_ratio * nn.num_units_in_each_layer[i])
                 for i in modify_layer_idxs]
  return change_num_units_in_layers(nn, modify_layer_idxs, modify_vals)

def _get_candidate_layers_for_modifying_num_units(nn, num_candidates='all'):
  """ Returns a set of candidate layers for modifying the number of units. """
  modifiable_layers = _get_directly_modifable_layer_idxs(nn)
  if num_candidates == 'all' or num_candidates >= len(modifiable_layers):
    return modifiable_layers
  else:
    np.random.shuffle(modifiable_layers)
    return modifiable_layers[:num_candidates]

def get_list_of_single_layer_modifiers(old_nn, inc_or_dec, num_layers_to_modify='all',
                                       change_frac=_DFLT_CHANGE_FRAC):
  """ Returns a list of primitives which change old_nn in one layer. """
  # Define a local function to obtain the modifier
  def _get_modifier(_ltm, _change_val):
    """ Gets a modifier with the current values of ltm and change_val. """
    return lambda nn: change_num_units_in_layers(nn, [_ltm], [_change_val])
  # The problem with doing ret.append(lamda nn ....(nn, ltm, change_val)) is that python
  # uses the current values of the variables ltm and change_val when calling - and this
  # turns out to be the last value in layers_to_modify. i.e. Python looks up the
  # variable name at the time the function is called, not when it is created.
  # See this:
  # https://stackoverflow.com/questions/10452770/python-lambdas-binding-to-local-values
  change_ratio = _get_change_ratio_from_change_frac(change_frac, inc_or_dec)
  layers_to_modify = _get_candidate_layers_for_modifying_num_units(old_nn,
                       num_layers_to_modify)
  ret = []
  for ltm in layers_to_modify:
    change_val = round(change_ratio * old_nn.num_units_in_each_layer[ltm])
    ret.append(_get_modifier(ltm, change_val))
  return ret

# Define the following for convenience
# Increase num units en masse
increase_en_masse_1_2 = lambda nn, *a: modify_several_layers(nn, 'increase', '1/2', *a)
increase_en_masse_2_2 = lambda nn, *a: modify_several_layers(nn, 'increase', '2/2', *a)
increase_en_masse_1_4 = lambda nn, *a: modify_several_layers(nn, 'increase', '1/4', *a)
increase_en_masse_2_4 = lambda nn, *a: modify_several_layers(nn, 'increase', '2/4', *a)
increase_en_masse_3_4 = lambda nn, *a: modify_several_layers(nn, 'increase', '3/4', *a)
increase_en_masse_4_4 = lambda nn, *a: modify_several_layers(nn, 'increase', '4/4', *a)
increase_en_masse_1_8 = lambda nn, *a: modify_several_layers(nn, 'increase', '1/8', *a)
increase_en_masse_2_8 = lambda nn, *a: modify_several_layers(nn, 'increase', '2/8', *a)
increase_en_masse_3_8 = lambda nn, *a: modify_several_layers(nn, 'increase', '3/8', *a)
increase_en_masse_4_8 = lambda nn, *a: modify_several_layers(nn, 'increase', '4/8', *a)
increase_en_masse_5_8 = lambda nn, *a: modify_several_layers(nn, 'increase', '5/8', *a)
increase_en_masse_6_8 = lambda nn, *a: modify_several_layers(nn, 'increase', '6/8', *a)
increase_en_masse_7_8 = lambda nn, *a: modify_several_layers(nn, 'increase', '7/8', *a)
increase_en_masse_8_8 = lambda nn, *a: modify_several_layers(nn, 'increase', '8/8', *a)
# Decrease num units en masse
decrease_en_masse_1_2 = lambda nn, *a: modify_several_layers(nn, 'decrease', '1/2', *a)
decrease_en_masse_2_2 = lambda nn, *a: modify_several_layers(nn, 'decrease', '2/2', *a)
decrease_en_masse_1_4 = lambda nn, *a: modify_several_layers(nn, 'decrease', '1/4', *a)
decrease_en_masse_2_4 = lambda nn, *a: modify_several_layers(nn, 'decrease', '2/4', *a)
decrease_en_masse_3_4 = lambda nn, *a: modify_several_layers(nn, 'decrease', '3/4', *a)
decrease_en_masse_4_4 = lambda nn, *a: modify_several_layers(nn, 'decrease', '4/4', *a)
decrease_en_masse_1_8 = lambda nn, *a: modify_several_layers(nn, 'decrease', '1/8', *a)
decrease_en_masse_2_8 = lambda nn, *a: modify_several_layers(nn, 'decrease', '2/8', *a)
decrease_en_masse_3_8 = lambda nn, *a: modify_several_layers(nn, 'decrease', '3/8', *a)
decrease_en_masse_4_8 = lambda nn, *a: modify_several_layers(nn, 'decrease', '4/8', *a)
decrease_en_masse_5_8 = lambda nn, *a: modify_several_layers(nn, 'decrease', '5/8', *a)
decrease_en_masse_6_8 = lambda nn, *a: modify_several_layers(nn, 'decrease', '6/8', *a)
decrease_en_masse_7_8 = lambda nn, *a: modify_several_layers(nn, 'decrease', '7/8', *a)
decrease_en_masse_8_8 = lambda nn, *a: modify_several_layers(nn, 'decrease', '8/8', *a)


def get_list_of_en_masse_change_primitives(nn, inc_or_dec='incdec'):
  """ Returns the list of primitives which changes the number of units en masse. """
  ret = []
  # Change 1/2
  if nn.num_internal_layers >= 4:
    if 'inc' in inc_or_dec:
      ret.extend([increase_en_masse_1_2, # increase
                  increase_en_masse_2_2,
                 ])
    if 'dec' in inc_or_dec:
      ret.extend([decrease_en_masse_1_2, # decrease
                  decrease_en_masse_2_2,
                 ])
  # Change 1/4
  if nn.num_internal_layers >= 8:
    if 'inc' in inc_or_dec:
      ret.extend([increase_en_masse_1_4, # increase
                  increase_en_masse_2_4,
                  increase_en_masse_3_4,
                  increase_en_masse_4_4,
                 ])
    if 'dec' in inc_or_dec:
      ret.extend([decrease_en_masse_1_4, # decrease
                  decrease_en_masse_2_4,
                  decrease_en_masse_3_4,
                  decrease_en_masse_4_4,
                 ])
  # Change 1/8
  if nn.num_internal_layers >= 16:
    if 'inc' in inc_or_dec:
      ret.extend([increase_en_masse_1_8, # increase
                  increase_en_masse_2_8,
                  increase_en_masse_3_8,
                  increase_en_masse_4_8,
                  increase_en_masse_5_8,
                  increase_en_masse_6_8,
                  increase_en_masse_7_8,
                  increase_en_masse_8_8,
                  ])
    if 'dec' in inc_or_dec:
      ret.extend([decrease_en_masse_1_8, # decrease
                  decrease_en_masse_2_8,
                  decrease_en_masse_3_8,
                  decrease_en_masse_4_8,
                  decrease_en_masse_5_8,
                  decrease_en_masse_6_8,
                  decrease_en_masse_7_8,
                  decrease_en_masse_8_8,
                  ])
  return ret

# A class to put it all together
# ========================================================================================
class NNModifier(object):
  """ A class for modifying a neural nework using many of the operations above. """

  def __init__(self, constraint_checker=None, options=None, reporter=None):
    """ Constructor. """
    self.reporter = get_reporter(reporter)
    self.constraint_checker = constraint_checker
    if options is None:
      options = load_options(nn_modifier_args)
    self.options = options

  def __call__(self, list_of_nns, num_modifications, num_steps_probs, max_num_steps=None,
               **kwargs):
    """ Takes a list of neural network nns and applies the library of changes to it to
        produce a list of modifications.
        num_steps_probs is a list of probabilities that indicate with what probability
        we want to use a certain number of steps.
    """
    # Preprocessing
    if not hasattr(list_of_nns, '__iter__'):
      list_of_nns = [list_of_nns]
    # determine num_steps_probs
    if num_steps_probs is None and isinstance(max_num_steps, (int, float)):
      num_steps_probs = np.ones((max_num_steps,))/float(max_num_steps)
    elif isinstance(num_steps_probs, (int, float)):
      num_steps_probs = np.zeros((num_steps_probs,))
      num_steps_probs[-1] = 1.0
    # Determine how many modifications per nn
    nn_idxs = list(range(len(list_of_nns)))
    if hasattr(num_modifications, '__iter__'):
      num_modifs_for_each_nn = num_modifications
    else:
      nn_choices_for_modif = np.random.choice(nn_idxs, num_modifications, replace=True)
      num_modifs_for_each_nn = [np.sum(nn_choices_for_modif == i) for i in nn_idxs]
    # Create a list
    ret = []
    for idx in nn_idxs:
      ret.extend(self.get_modifications_for_a_single_nn(list_of_nns[idx],
        num_modifs_for_each_nn[idx], num_steps_probs, **kwargs))
    return ret

  def get_modifications_for_a_single_nn(self, nn, num_modifications, num_steps_probs,
                                        **kwargs):
    """ Takes a neural network nn and applies the library of changes to it to produce
        a list of modifiers.
        num_steps_probs is a list of probabilities that indicate with what probability
        we want to use a certain number of steps.
    """
    max_num_steps = len(num_steps_probs)
    num_step_choices = np.random.choice(list(range(max_num_steps)), num_modifications,
                                        replace=True, p=num_steps_probs)
    num_modifs_per_step = [np.sum(num_step_choices == i) for i in range(max_num_steps)]
    ret = []
    for num_step_val_minus1, num_modifs_this_step in enumerate(num_modifs_per_step):
      if num_modifs_this_step == 0:
        continue
      new_nns = self.get_multi_step_modifications(nn, num_step_val_minus1 + 1,
                                                  num_modifs_this_step, **kwargs)
      ret.extend(new_nns)
    return ret

  @classmethod
  def _get_num_modifications(cls, list_of_opers, passed_num_modifs, dflt_num_modifs):
    """ Returns the number of modifications depending on the passed value. """
    if passed_num_modifs == 'all':
      return len(list_of_opers)
    elif passed_num_modifs is None:
      return dflt_num_modifs
    elif passed_num_modifs >= 0:
      return passed_num_modifs
    else:
      raise ValueError('num_*_step_modifications should be \'all\', ' +
                       'None or a positive integer. ')

  def _is_a_valid_network(self, nn):
    """ Returns true if it is a valid network. """
    if nn is None:
      return False
    elif self.constraint_checker is None:
      return True
    else:
      return self.constraint_checker(nn)

  def get_primitives_grouped_by_type(self, nn, types_of_primitives=None):
    """ Returns the list of primitives grouped by type. """
    # pylint: disable=bare-except
    # doing a bare except only for testing. Will remove soon.
    types_of_primitives = types_of_primitives if types_of_primitives is not None \
                          else _PRIMITIVE_PROB_MASSES.keys()
    ret = {}
    primitive_type_prob_masses = {}
    for top in types_of_primitives:
      top_is_of_known_type = True
      try:
        if top == 'inc_single':
          ret[top] = get_list_of_single_layer_modifiers(nn, 'increase',
                       self.options.spawn_single_inc_num_units,
                       self.options.single_inc_change_frac)
        elif top == 'dec_single':
          ret[top] = get_list_of_single_layer_modifiers(nn, 'decrease',
                       self.options.spawn_single_dec_num_units,
                       self.options.single_dec_change_frac)
        elif top == 'inc_en_masse':
          ret[top] = get_list_of_en_masse_change_primitives(nn, 'inc')
        elif top == 'dec_en_masse':
          ret[top] = get_list_of_en_masse_change_primitives(nn, 'dec')
        elif top == 'swap_layer':
          ret[top] = get_list_of_swap_layer_modifiers(nn)
        elif top == 'wedge_layer':
          ret[top] = get_list_of_wedge_layer_modifiers(nn)
        elif top == 'remove_layer':
          ret[top] = get_list_of_remove_layer_modifiers(nn)
        elif top == 'branch':
          ret[top] = get_list_of_branching_modifiers(nn)
        elif top == 'skip':
          ret[top] = get_list_of_skipping_modifiers(nn)
        else:
          top_is_of_known_type = False
      except:
        pass
      if not top_is_of_known_type:
        raise ValueError('All values in types_of_primitives should be in %s. Given %s'%(
                         str(_PRIMITIVE_PROB_MASSES.keys()), top))
      # Finally also add the probability mass
      primitive_type_prob_masses[top] = _PRIMITIVE_PROB_MASSES[top]
    return ret, primitive_type_prob_masses

  def get_single_step_modifications(self, nn, num_single_step_modifications='all',
                                    **kwargs):
    """ Returns a list of new neural networks which have undergone a single modification
        from nn. """
    prims_by_type, type_prob_masses = self.get_primitives_grouped_by_type(nn, **kwargs)
    groups = prims_by_type.keys()
    if num_single_step_modifications == 'all':
      prims_by_group = prims_by_type.values()
      modifiers = [elem for group_prims in prims_by_group for elem in group_prims]
      num_single_step_modifications = len(modifiers)
    else:
      prob_masses = np.array([type_prob_masses[key] for key in groups])
      prob_masses = prob_masses/prob_masses.sum()
      modif_groups = np.random.choice(groups, max(2*num_single_step_modifications, 20),
                                      p=prob_masses)
      # Shuffle each list
      for grp in groups:
        np.random.shuffle(prims_by_type[grp])
      # Now create a list of modifiers
      modifiers = []
      for grp in modif_groups:
        if len(prims_by_type[grp]) == 0:
          continue
        else:
          modifiers.append(prims_by_type[grp].pop())
    # Now go through each modifier and make changes.
    ret = []
    for modif in modifiers:
      curr_val = modif(nn)
      if self._is_a_valid_network(curr_val):
        ret.append(curr_val)
        if len(ret) >= num_single_step_modifications:
          break
    if len(ret) < 0.25 * num_single_step_modifications:
      self.reporter.writeln(('Consider reducing constraints. %d/%d modifications ' +
                             'accepted.')%(len(ret), num_single_step_modifications))
    return ret

  def get_multi_step_modifications(self, nn, num_steps, num_multi_step_modifications,
                                   **kwargs):
    """ Returns a list of new neural networks which have undergone num_steps
        modifications.
    """
    single_step_modifs = self.get_single_step_modifications(nn,
                           num_multi_step_modifications, **kwargs)
    if num_steps == 1:
      return single_step_modifs
    else:
      ret = []
      for new_nn in single_step_modifs:
        result = self.get_multi_step_modifications(new_nn, num_steps-1, 1, **kwargs)
        assert len(result) <= 1
        if len(result) > 0:
          ret.append(result[0])
        else:
          ret.append(new_nn)
      return ret


# APIs =================================================================================
# An API to simply return an operation
def get_nn_modifier_from_args(constraint_checker, dflt_num_steps_probs=None,
                              dflt_max_num_steps=None, options=None, reporter=None):
  """ Returns a function which can be used to modify neural networks. """
  # A modifier which works with probabilities
  def _get_modifier_with_probs(_nn_modifier, _num_steps_probs):
    """ Returns a modifier with probs. """
    _num_steps_probs = np.array(_num_steps_probs)
    _num_steps_probs = _num_steps_probs/_num_steps_probs.sum()
    return lambda list_of_nns, num_modifications, **kwargs: _nn_modifier(list_of_nns,
                                      num_modifications, num_steps_probs=_num_steps_probs,
                                      max_num_steps=None, **kwargs)
  # A modifier which works with max_num_steps
  def _get_modifier_with_steps(_nn_modifier, _max_num_steps):
    """ Returns a modifier with max_num_steps. """
    return lambda list_of_nns, num_modifications, **kwargs: _nn_modifier(list_of_nns,
                                      num_modifications, num_steps_probs=None,
                                      max_num_steps=_max_num_steps, **kwargs)
  # Construct a Modifier
  nn_modifier = NNModifier(constraint_checker, options, reporter)
  if dflt_num_steps_probs is not None:
    return _get_modifier_with_probs(nn_modifier, dflt_num_steps_probs)
  elif dflt_max_num_steps is not None:
    return _get_modifier_with_steps(nn_modifier, dflt_max_num_steps)
  else:
    return nn_modifier

# An API to return a mutation for a single nn
def get_single_nn_mutation_op(dom, dflt_num_steps_probs=None, dflt_max_num_steps=None,
                              options=None, reporter=None):
  """ Returns a mutation operator that will take a single nn and return a mutated version
      of the NN.
  """
  # A modifier which works with probabilities
  def _get_single_mutation_from_probs(_nn_modifier, _num_steps_probs):
    """ Returns a single mutation operator with probs. """
    _num_steps_probs = np.array(_num_steps_probs)
    _num_steps_probs = _num_steps_probs/_num_steps_probs.sum()
    return lambda nn: \
      _nn_modifier.get_modifications_for_a_single_nn(nn, 1, _num_steps_probs)[0]
  # Create modifier and return
  nn_modifier = NNModifier(dom.constraint_checker, options, reporter)
  if dflt_num_steps_probs is not None:
    return _get_single_mutation_from_probs(nn_modifier, dflt_num_steps_probs)
  elif dflt_max_num_steps is not None:
    num_steps_probs = np.zeros((dflt_max_num_steps, ))
    num_steps_probs[-1] = 1.0
    return _get_single_mutation_from_probs(nn_modifier, num_steps_probs)
  else:
    raise ValueError('Specify either dflt_num_steps_probs or dflt_max_num_steps.')




"""
  Adapted from pgmpy library: https://github.com/pgmpy/pgmpy
  -- kvysyara@andrew.cmu.edu
"""

# Local imports
from .check_functions import _check_1d_array_object, _check_length_equal

class BaseSimulateHamiltonianDynamics(object):
  """
  Base class for proposing new values of position and momentum
  by simulating Hamiltonian Dynamics.

  Classes inheriting this base class can be passed as an argument for
  simulate_dynamics in inference algorithms.

  Parameters
  ----------
  model : Model for which DiscretizeTime object is initialized

  position : A 1d array like object (numpy.ndarray or list)
      Vector representing values of parameter position( or X)

  momentum: A 1d array like object (numpy.ndarray or list)
      Vector representing the proposed value for momentum (velocity)

  stepsize: Float
      stepsize for the simulating dynamics
  """

  def __init__(self, model, position, momentum, stepsize):

    position = _check_1d_array_object(position, 'position')

    momentum = _check_1d_array_object(momentum, 'momentum')

    _check_length_equal(position, momentum, 'position', 'momentum')

    self.position = position
    self.momentum = momentum
    self.stepsize = stepsize
    self.model = model

    # new_position is the new proposed position, new_momentum is the new
    # proposed momentum, new_grad_lop is the value of grad log at new_position
    self.new_position = self.new_momentum = self.new_grad_logp = None

  def get_proposed_values(self):
    """
    Returns new proposed values of position and momentum
    """
    return self.new_position, self.new_momentum, self.new_grad_logp


class LeapFrog(BaseSimulateHamiltonianDynamics):
  """
  Class for simulating hamiltonian dynamics using leapfrog method

  Parameters
  ----------
  model : An instance of pgmpy.models
      Model for which DiscretizeTime object is initialized

  position : A 1d array like object (numpy.ndarray or list)
      Vector representing values of parameter position( or X)

  momentum: A 1d array like object (numpy.ndarray or list)
      Vector representing the proposed value for momentum (velocity)

  stepsize: Float
      stepsize for the simulating dynamics
  """

  def __init__(self, model, position, momentum, stepsize):

    BaseSimulateHamiltonianDynamics.__init__(self, model, position, momentum, stepsize)

    self.new_position, self.new_momentum, self.new_grad_logp = self._get_proposed_values()

  def _get_proposed_values(self):
    """
    Method to perform time splitting using leapfrog
    """
    # Take half step in time for updating momentum
    momentum_bar = self.momentum + 0.5*self.stepsize*self.model.grad_logp(self.position)

    # Take full step in time for updating position position
    position_bar = self.position + self.stepsize * momentum_bar

    grad_log = self.model.grad_logp(position_bar)

    # Take remaining half step in time for updating momentum
    momentum_bar = momentum_bar + 0.5 * self.stepsize * grad_log

    return position_bar, momentum_bar, grad_log



"""
  A collection of utilities needed for the APIs.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=maybe-no-member
# pylint: disable=no-member


from argparse import Namespace
# Local imports
from ..exd.cp_domain_utils import load_config_file, \
                                get_processed_func_from_raw_func_for_cp_domain_fidelity, \
                                get_processed_func_from_raw_func_for_cp_domain
from ..exd.domains import EuclideanDomain
from ..exd.worker_manager import SyntheticWorkerManager, RealWorkerManager
from ..opt.ga_optimiser import ga_opt_args
from ..opt.gp_bandit import get_all_euc_gp_bandit_args, \
                            get_all_cp_gp_bandit_args, get_all_mf_euc_gp_bandit_args, \
                            get_all_mf_cp_gp_bandit_args
from ..opt.random_optimiser import euclidean_random_optimiser_args, \
                                   mf_euclidean_random_optimiser_args, \
                                   cp_random_optimiser_args, mf_cp_random_optimiser_args
from ..opt.multiobjective_gp_bandit import get_all_euc_moo_gp_bandit_args, \
                                           get_all_cp_moo_gp_bandit_args
from ..opt.random_multiobjective_optimiser import \
              euclidean_random_multiobjective_optimiser_args, \
              cp_random_multiobjective_optimiser_args
from ..utils.option_handler import load_options


def get_worker_manager_from_capital_type(capital_type, num_workers=1, tmp_dir=None):
  """ Get worker manager. """
  if capital_type in ['return_value', 'num_evals']:
    return SyntheticWorkerManager(num_workers=num_workers)
  elif capital_type == 'realtime':
    if tmp_dir is None:
      from datetime import datetime
      tmp_dir = './tmp_%s'%(datetime.now().strftime('%m%d_%H%M%S'))
    return RealWorkerManager(worker_ids=num_workers, tmp_dir=tmp_dir)
  else:
    raise ValueError('Unknown Capital Type: %s.'%(capital_type))


def _raise_load_options_not_supported_error(method, prob, domain_type, capital_type):
  """ Raises an exception if the combination of method/prob/domain are not supported. """
  # problem
  prob_str_dict = {'opt': 'optimisation', 'moo': 'multi-objective optimisation',
                   'mfopt': 'multi-fidelity optimisation'}
  if prob not in prob_str_dict.keys():
    raise ValueError('Unknown problem %s.'%(prob))
  prob_str = prob_str_dict[prob]
  # method
  meth_str_dict = {'bo': 'Bayesian-Optimisation', 'ea': 'Evolutionary-Algorithms',
                   'ga': 'Evolutionary-Algorithms', 'direct': 'DiRect', 'pdoo': 'PDOO'}
  if method not in meth_str_dict.keys():
    raise ValueError('Unknown method %s.'%(method))
  meth_str = meth_str_dict[method]
  # domain
  if domain_type not in ['euclidean', 'cartesian_product']:
    dom_err_msg = 'We do not support directly optimising on %s spaces. Consider ' + \
                  'wrapping your domain with a CartesianProduct Domain. See ' + \
                  'dragonfly/exd/domains.py.'
    raise ValueError(dom_err_msg)
  err_msg = ('Draogonfly not support %s problems with %s on %s domains with capital ' +
             'type %s.')%(prob_str, meth_str, domain_type, capital_type)
  raise ValueError(err_msg)


def load_options_for_method(method, prob, domain, capital_type, options=None):
  """ Loads and returns the options for the problem. """
  # pylint: disable=too-many-branches
  # Both ea and ga can be used to specify evolutionary algorithms
  method = 'ga' if method == 'ea' else method
  # Now load options according to the case
  case_sel = (method, prob, domain.get_type())
  if case_sel == ('ga', 'opt', 'cartesian_product'):
    opt_options = load_options(ga_opt_args)
  elif case_sel == ('bo', 'opt', 'euclidean'):
    opt_options = load_options(get_all_euc_gp_bandit_args())
  elif case_sel == ('bo', 'opt', 'cartesian_product'):
    opt_options = load_options(get_all_cp_gp_bandit_args())
  elif (case_sel == ('direct', 'opt', 'euclidean') or \
        case_sel == ('pdoo', 'opt', 'euclidean')) and \
       capital_type in ['return_value', 'num_evals']:
    opt_options = Namespace()
  elif case_sel == ('rand', 'opt', 'euclidean'):
    opt_options = load_options(euclidean_random_optimiser_args)
  elif case_sel == ('rand', 'opt', 'cartesian_product'):
    opt_options = load_options(cp_random_optimiser_args)
  elif case_sel == ('rand', 'mfopt', 'euclidean'):
    opt_options = load_options(mf_euclidean_random_optimiser_args)
  elif case_sel == ('rand', 'mfopt', 'cartesian_product'):
    opt_options = load_options(mf_cp_random_optimiser_args)
  elif case_sel == ('bo', 'mfopt', 'euclidean'):
    opt_options = load_options(get_all_mf_euc_gp_bandit_args())
  elif case_sel == ('bo', 'mfopt', 'cartesian_product'):
    opt_options = load_options(get_all_mf_cp_gp_bandit_args())
  elif case_sel == ('bo', 'moo', 'euclidean'):
    opt_options = load_options(get_all_euc_moo_gp_bandit_args())
  elif case_sel == ('bo', 'moo', 'cartesian_product'):
    opt_options = load_options(get_all_cp_moo_gp_bandit_args())
  elif case_sel == ('rand', 'moo', 'euclidean'):
    opt_options = load_options(euclidean_random_multiobjective_optimiser_args)
  elif case_sel == ('rand', 'moo', 'cartesian_product'):
    opt_options = load_options(cp_random_multiobjective_optimiser_args)
  else:
    _raise_load_options_not_supported_error(method, prob, domain.get_type(), capital_type)
  # Now look at any options that have been passed in already and add them to opt_options
  options = Namespace() if options is None else options
  for attr in vars(options):
    setattr(opt_options, attr, getattr(options, attr))
  # Also add capital type
  opt_options.capital_type = 'return_value' if capital_type == 'num_evals' \
                             else capital_type
  return opt_options


def preprocess_multifidelity_arguments(fidel_space, domain, funcs, fidel_cost_func,
                                       fidel_to_opt, config):
  """ Preprocess fidel_space, domain arguments and configuration file. """
  # Preprocess config argument
  converted_cp_to_euclidean = False
  if isinstance(config, str):
    config = load_config_file(config)
  if fidel_space is None:
    fidel_space = config.fidel_space
  if domain is None:
    domain = config.domain
  if fidel_to_opt is None:
    fidel_to_opt = config.fidel_to_opt
  # The function
  if config is not None:
    proc_funcs = [get_processed_func_from_raw_func_for_cp_domain_fidelity(f, config)
                  for f in funcs]
    proc_fidel_cost_func = get_processed_func_from_raw_func_for_cp_domain(fidel_cost_func,
      config.fidel_space, config.fidel_space_orderings.index_ordering,
      config.fidel_space_orderings.dim_ordering)
  else:
    proc_funcs = funcs
    proc_fidel_cost_func = fidel_cost_func
  ret_funcs = proc_funcs
  ret_fidel_cost_func = proc_fidel_cost_func
  # Preprocess domain argument
  if isinstance(fidel_space, (list, tuple)) and isinstance(domain, (list, tuple)):
    domain = EuclideanDomain(domain)
    fidel_space = EuclideanDomain(fidel_space)
  elif fidel_space.get_type() == 'euclidean' and domain.get_type() == 'euclidean':
    pass
  elif fidel_space.get_type() == 'cartesian_product' and \
       domain.get_type() == 'cartesian_product':
    if fidel_space.num_domains == 1 and \
         fidel_space.list_of_domains[0].get_type() == 'euclidean' and \
       domain.num_domains == 1 and domain.list_of_domains[0].get_type() == 'euclidean':
      # Change the fidelity space
      fidel_space = fidel_space.list_of_domains[0]
      config.fidel_space_orderings.dim_ordering = \
        config.fidel_space_orderings.dim_ordering[0]
      config.fidel_space_orderings.index_ordering = \
        config.fidel_space_orderings.index_ordering[0]
      config.fidel_space_orderings.kernel_ordering = \
        config.fidel_space_orderings.kernel_ordering[0]
      config.fidel_space_orderings.name_ordering = \
        config.fidel_space_orderings.name_ordering[0]
      config.fidel_to_opt = config.fidel_to_opt[0]
      fidel_to_opt = fidel_to_opt[0]
      # Change the domain
      domain = domain.list_of_domains[0]
      config.domain_orderings.dim_ordering = config.domain_orderings.dim_ordering[0]
      config.domain_orderings.index_ordering = config.domain_orderings.index_ordering[0]
      config.domain_orderings.kernel_ordering = config.domain_orderings.kernel_ordering[0]
      config.domain_orderings.name_ordering = config.domain_orderings.name_ordering[0]
      # Add to config
      config.fidel_space = fidel_space
      config.domain = domain
      converted_cp_to_euclidean = True
      # Functions
      def _get_ret_func_from_proc_func_for_euc_domains(_proc_func):
        """ Get function to return. """
        return lambda z, x: _proc_func([z], [x])
      ret_funcs = [_get_ret_func_from_proc_func_for_euc_domains(pf) for pf in proc_funcs]
      ret_fidel_cost_func = lambda z: proc_fidel_cost_func([z])
  else:
    raise ValueError('fidel_space and domain should be either both instances of ' +
                     'EuclideanDomain or both CartesianProductDomain.')
  return (fidel_space, domain, ret_funcs, ret_fidel_cost_func, fidel_to_opt, config,
          converted_cp_to_euclidean)


def preprocess_arguments(domain, funcs, config):
  """ Preprocess domain arguments and configuration file. """
  # Preprocess config argument
  converted_cp_to_euclidean = False
  if isinstance(config, str):
    config = load_config_file(config)
  if domain is None:
    domain = config.domain
  # The function
  if config is not None:
    proc_funcs = [get_processed_func_from_raw_func_for_cp_domain(f, config.domain,
      config.domain_orderings.index_ordering, config.domain_orderings.dim_ordering)
      for f in funcs]
  else:
    proc_funcs = funcs
  ret_funcs = proc_funcs
  # Preprocess domain argument
  if isinstance(domain, (list, tuple)):
    domain = EuclideanDomain(domain)
  elif domain.get_type() == 'euclidean':
    pass
  elif domain.get_type() == 'cartesian_product':
    if domain.num_domains == 1 and domain.list_of_domains[0].get_type() == 'euclidean':
      domain = domain.list_of_domains[0]
      config.domain_orderings.dim_ordering = config.domain_orderings.dim_ordering[0]
      config.domain_orderings.index_ordering = config.domain_orderings.index_ordering[0]
      config.domain_orderings.kernel_ordering = config.domain_orderings.kernel_ordering[0]
      config.domain_orderings.name_ordering = config.domain_orderings.name_ordering[0]
      config.domain = domain
      converted_cp_to_euclidean = True
      # The function
      def _get_ret_func_from_proc_func_for_euc_domains(_proc_func):
        """ Get function to return. """
        return lambda x: _proc_func([x])
      ret_funcs = [_get_ret_func_from_proc_func_for_euc_domains(pf) for pf in proc_funcs]
  else:
    raise ValueError('domain should be an instance of EuclideanDomain or ' +
                     'CartesianProductDomain.')
  return domain, ret_funcs, config, converted_cp_to_euclidean


def post_process_history_for_minimisation(history):
  """ Post processes history for minimisation. """
  # Negate past queries
  history.query_vals = [-qv for qv in history.query_vals]
  history.curr_opt_vals = [-cov for cov in history.curr_opt_vals]
  history.curr_true_opt_vals = [-cov for cov in history.curr_true_opt_vals]
  return history




"""
  Unit tests for mf_gp.py and the MFGP objects in euclidean_gp.py.
  -- kandasamy@cs.cmu.edu
"""
from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=no-member
# pylint: disable=maybe-no-member

from argparse import Namespace
from copy import deepcopy
import numpy as np
# Local
from . import kernel
from .unittest_euclidean_gp import fit_se_gp_with_dataset, fit_matern_gp_with_dataset, \
                                     fit_poly_gp_with_dataset
from . import euclidean_gp
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.general_utils import compute_average_sq_prediction_error
from ..utils.option_handler import load_options


# Some wrappers for functions in gp.unittest_gp_instances --------------------------------
def fit_gp_with_dataset(dataset, kernel_type='se', *args, **kwargs):
  """ A wrapper to fit a gp with the given dataset. """
  if kernel_type == 'se':
    return fit_se_gp_with_dataset(dataset, *args, **kwargs)
  elif kernel_type == 'matern':
    return fit_matern_gp_with_dataset(dataset, *args, **kwargs)
  elif kernel_type == 'poly':
    return fit_poly_gp_with_dataset(dataset, *args, **kwargs)
  else:
    raise ValueError('Unknown kernel type %s.'%(kernel_type))

# Functions to create data ---------------------------------------------------------------
def _get_mf_gp_options(tune_noise):
  """ Gets the options for the dataset. """
  options = load_options(euclidean_gp.euclidean_mf_gp_args)
  options.noise_var_type = 'tune' if tune_noise else options.noise_var_type
  return options

def gen_data_from_func(fzx, N, dim_z, dim_x):
  """ Generates data from the function. """
  Z = np.random.random((N, dim_z))
  X = np.random.random((N, dim_x))
  Y = fzx(Z, X)
  return Z, X, Y

def _gen_datasets_from_func(fzx, N, dim_z, dim_x):
  """ Generates train and test datasets from the function. """
  Z_tr, X_tr, Y_tr = gen_data_from_func(fzx, N, dim_z, dim_x)
  ZX_tr = np.concatenate((Z_tr, X_tr), axis=1)
  Z_te, X_te, Y_te = gen_data_from_func(fzx, 2 * N, dim_z, dim_x)
  ZX_te = np.concatenate((Z_te, X_te), axis=1)
  return Z_tr, X_tr, Y_tr, ZX_tr, Z_te, X_te, Y_te, ZX_te

def gen_mf_gp_test_data():
  """ Generates test data. """
  # pylint: disable=too-many-locals
  # dataset 1
  fzx = lambda z, x: (x**2).sum(axis=1) + (z**2).sum(axis=1)
  dim_z = 2
  dim_x = 3
  N = 20
  Z_tr, X_tr, Y_tr, ZX_tr, Z_te, X_te, Y_te, ZX_te = _gen_datasets_from_func(
    fzx, N, dim_z, dim_x)
  fidel_kernel = kernel.SEKernel(dim=dim_z, scale=1, dim_bandwidths=0.5)
  domain_kernel = kernel.SEKernel(dim=dim_x, scale=1, dim_bandwidths=0.5)
  kernel_scale = 2
  tune_noise = True
  dataset_1 = Namespace(Z_tr=Z_tr, X_tr=X_tr, Y_tr=Y_tr, ZX_tr=ZX_tr,
                        Z_te=Z_te, X_te=X_te, Y_te=Y_te, ZX_te=ZX_te,
                        fidel_kernel=fidel_kernel, domain_kernel=domain_kernel,
                        kernel_scale=kernel_scale, tune_noise=tune_noise)
  # dataset 2
  fx = lambda x: -70 * (x + 0.01) * (x - 0.31) * (x + 0.51) * (x - 0.71) * (x - 0.98)
  fzx = lambda z, x: (np.exp((z - 0.8)**2) * fx(x)).sum(axis=1)
  N = 100
  Z_tr, X_tr, Y_tr, ZX_tr, Z_te, X_te, Y_te, ZX_te = _gen_datasets_from_func(fzx, N, 1, 1)
  fidel_kernel = kernel.SEKernel(dim=1, scale=1, dim_bandwidths=1.0)
  domain_kernel = kernel.SEKernel(dim=1, scale=1, dim_bandwidths=0.25)
  kernel_scale = 2
  tune_noise = True
  dataset_2 = Namespace(Z_tr=Z_tr, X_tr=X_tr, Y_tr=Y_tr, ZX_tr=ZX_tr,
                        Z_te=Z_te, X_te=X_te, Y_te=Y_te, ZX_te=ZX_te,
                        fidel_kernel=fidel_kernel, domain_kernel=domain_kernel,
                        kernel_scale=kernel_scale, tune_noise=tune_noise)
  # return all datasets
  return [dataset_1, dataset_2]

def get_init_and_post_gp(fidel_dim, domain_dim, num_data, kernel_scale, dim_bw_power=0.5):
  """ Generates a GP and data, constructs posterior and returns everything. """
  # pylint: disable=too-many-locals
  kernel_bw_scaling = float(fidel_dim + domain_dim) ** dim_bw_power
  fidel_kernel = kernel.SEKernel(fidel_dim, 1,
                    (1 + np.random.random(fidel_dim)) * kernel_bw_scaling)
  domain_kernel = kernel.SEKernel(domain_dim, 1,
                    (0.1 + 0.2*np.random.random(domain_dim)) * kernel_bw_scaling)
  mean_func_const_val = (2 + np.random.random()) * (1 + np.random.random())
  mean_func = lambda x: np.array([mean_func_const_val] * len(x))
  noise_var = 0.05 * np.random.random()
  Z_init = np.zeros((0, fidel_dim))
  X_init = np.zeros((0, domain_dim))
  Y_init = np.zeros((0))
  init_gp = euclidean_gp.EuclideanMFGP(Z_init, X_init, Y_init, None, kernel_scale,
                       fidel_kernel, domain_kernel, mean_func, noise_var)
  # Now construct the data
  post_gp = deepcopy(init_gp)
  Z_data = np.random.random((num_data, fidel_dim))
  X_data = np.random.random((num_data, domain_dim))
  Y_wo_noise = post_gp.draw_mf_samples(1, Z_data, X_data).ravel()
  Y_data = Y_wo_noise + np.random.normal(0, np.sqrt(noise_var), Y_wo_noise.shape)
  post_gp.add_mf_data_multiple(Z_data, X_data, Y_data)
  # Put everything in a namespace
  ret = Namespace(init_gp=init_gp, post_gp=post_gp, Z_data=Z_data, X_data=X_data,
    Y_data=Y_data, fidel_kernel=fidel_kernel, domain_kernel=domain_kernel,
    mean_func=mean_func, noise_var=noise_var, kernel_scale=kernel_scale,
    fidel_dim=fidel_dim, domain_dim=domain_dim, num_data=num_data)
  return ret

def gen_mf_gp_instances():
  """ Generates some MF GP instances. """
  # pylint: disable=star-args
  # The following list of lists maintains each problem instance in the following
  # order. (fidel_dim, domain_dim, num_data, kernel_scale)
  data = [[1, 1, 5, 1], [2, 4, 15, 2], [5, 10, 100, 2], [3, 20, 200, 4]]
  return [get_init_and_post_gp(*d) for d in data]


# Functions to build GPs -----------------------------------------------------------------
def build_mfgp_with_dataset(dataset, *_):
  """ Builds a MF GP by using some reasonable values for the parameters. """
  mean_func = lambda x: np.array([np.median(dataset.Y_tr)] * len(x))
  noise_var = (dataset.Y_tr.std()**2)/20
  return euclidean_gp.EuclideanMFGP(dataset.Z_tr, dataset.X_tr, dataset.Y_tr, None,
                             dataset.kernel_scale, dataset.fidel_kernel,
                             dataset.domain_kernel, mean_func, noise_var)

def fit_simple_gp_with_dataset(dataset, kernel_type, *args, **kwargs):
  """ Builds a simple GP with the dataset. """
  return fit_gp_with_dataset([dataset.ZX_tr, dataset.Y_tr], kernel_type, *args, **kwargs)

def fit_mfgp_with_dataset_on_fidel_domain_kernels(dataset, fidel_kernel_type,
                                                  domain_kernel_type):
  """ Fits a GP with the dataset using different fidelity and domain kernels. """
  options = _get_mf_gp_options(dataset.tune_noise)
  options.fidel_kernel_type = fidel_kernel_type
  options.domain_kernel_type = domain_kernel_type
  options.mean_func_type = 'median'
  ret_fit_gp = (euclidean_gp.EuclideanMFGPFitter(dataset.Z_tr, dataset.X_tr, dataset.Y_tr,
                options=options)).fit_gp()
  assert ret_fit_gp[0] == 'fitted_gp'
  return ret_fit_gp[1]

def fit_mfgp_with_dataset(dataset, kernel_type):
  """ Fits a GP with the dataset. """
  return fit_mfgp_with_dataset_on_fidel_domain_kernels(dataset, kernel_type, kernel_type)

def fit_mfgp_with_expdecay_fidel_kernel(dataset, domain_kernel_type):
  """ Fits a GP with the dataset using an expdecay kernel for the fidelity space. """
  return fit_mfgp_with_dataset_on_fidel_domain_kernels(dataset, 'expdecay',
                                                       domain_kernel_type)

def fit_add_mfgp_with_dataset(dataset, kernel_type):
  """ Fits a GP with the dataset. """
  options = _get_mf_gp_options(dataset.tune_noise)
  options.fidel_kernel_type = kernel_type
  options.domain_kernel_type = kernel_type
  options.domain_use_additive_gp = True
  options.mean_func_type = 'median'
  options.domain_matern_nu = 2.5
  options.fidel_matern_nu = 2.5
  options.ml_hp_tune_opt = 'direct'
  ret_fit_gp = (euclidean_gp.EuclideanMFGPFitter(dataset.Z_tr, dataset.X_tr, dataset.Y_tr,
                     options=options)).fit_gp()
  assert ret_fit_gp[0] == 'fitted_gp'
  return ret_fit_gp[1]


# Test cases -----------------------------------------------------------------------------
class EuclideanMFGPTestCase(BaseTestClass):
  """ Unit tests for the MF GP. """
  # pylint: disable=too-many-locals

  def setUp(self):
    """ Set up for the tests. """
    self.datasets = gen_mf_gp_test_data()

  def test_eval_at_fidel(self):
    """ Tests eval at fidel. """
    self.report('MFGP.eval_at_fidel vs GP.eval.')
    ds = self.datasets[0]
    curr_gp = build_mfgp_with_dataset(ds)
    curr_pred, curr_std = curr_gp.eval_at_fidel(ds.Z_te, ds.X_te, uncert_form='std')
    alt_pred, alt_std = curr_gp.eval(np.concatenate((ds.Z_te, ds.X_te), axis=1),
                                     uncert_form='std')
    assert np.linalg.norm(curr_pred - alt_pred) < 1e-5
    assert np.linalg.norm(curr_std - alt_std) < 1e-5

  def test_eval(self):
    """ Tests the evaluation. """
    self.report('MFGP.eval_at_fidel: Probabilistic test, might fail sometimes.')
    num_successes = 0
    for ds in self.datasets:
      curr_gp = build_mfgp_with_dataset(ds)
      curr_pred, _ = curr_gp.eval_at_fidel(ds.Z_te, ds.X_te)
      curr_err = compute_average_sq_prediction_error(ds.Y_te, curr_pred)
      const_err = compute_average_sq_prediction_error(ds.Y_te, ds.Y_tr.mean())
      success = curr_err < const_err
      self.report(('(N,DZ,DX)=' + str(ds.Z_tr.shape + (ds.X_tr.shape[1],)) +
                   ':: MFGP-err= ' + str(curr_err) + ',   Const-err= ' + str(const_err) +
                   ',  success=' + str(success)), 'test_result')
      num_successes += int(success)
    assert num_successes > 0.6 *len(self.datasets)

  def test_compute_log_marginal_likelihood(self):
    """ Tests compute_log_marginal_likelihood. Does not test for accurate implementation.
        Only tests if the function runs without runtime errors. """
    self.report('MFGP.compute_log_marginal_likelihood: ** Runtime test errors only **')
    for ds in self.datasets:
      curr_gp = build_mfgp_with_dataset(ds)
      lml = curr_gp.compute_log_marginal_likelihood()
      self.report(('(N,DZ,DX)=' + str(ds.Z_tr.shape + (ds.X_tr.shape[1],)) +
                   ':: MFGP-lml= ' + str(lml)), 'test_result')


# Some common functionality for all MF Euclidean Fitter tests -------------------------
class EuclideanMFGPFitterBaseTestClass(object):
  """ Some common functionality for all MF Euclidean Fitter tests. """
  # pylint: disable=no-member

  def setUp(self):
    """ Set up for the tests. """
    self.datasets = gen_mf_gp_test_data()

  def _test_marginal_likelihood(self, get_gp_func1, get_gp_func2, descr1, descr2,
                                compare_fits=True):
    """ Test for marginal likelihood. """
    num_lml_successes = 0
    num_err_successes = 0
    num_experiments = 0
    for ds_idx, ds in enumerate(self.datasets):
      for kernel_type in ['se', 'matern']:
        gp1 = get_gp_func1(ds, kernel_type)
        gp2 = get_gp_func2(ds, kernel_type)
        lml1 = gp1.compute_log_marginal_likelihood()
        lml2 = gp2.compute_log_marginal_likelihood()
        preds1, _ = gp1.eval(ds.ZX_te)
        preds2, _ = gp2.eval(ds.ZX_te)
        err1 = compute_average_sq_prediction_error(ds.Y_te, preds1)
        err2 = compute_average_sq_prediction_error(ds.Y_te, preds2)
        lml_success = lml1 <= lml2
        err_success = err1 >= err2
        self.report('%d. (N,DZ,DX)= %s, kern=%s %s-lml=%0.4f, %s-lml=%0.4f, succ=%d'\
         %(ds_idx+1, str(ds.Z_tr.shape + (ds.X_tr.shape[1],)), kernel_type, descr1, lml1,
           descr2, lml2, lml_success), 'test_result')
        self.report('  %s-err = %0.4f, %s-err = %0.4f, err_succ=%d.'%(
          descr1, err1, descr2, err2, err_success), 'test_result')
        self.report('  %s GP :: %s'%(descr1, str(gp1)), 'test_result')
        self.report('  %s GP:: %s'%(descr2, str(gp2)), 'test_result')
        num_lml_successes += lml_success
        num_err_successes += err_success
        num_experiments += 1
    if compare_fits:
      assert num_lml_successes > 0.6 * num_experiments
      assert num_err_successes > 0.6 * num_experiments


# A Test for EuclideanMFGP (without additive models) -----------------------------
# class EuclideanMFGPFitterTestCase(EuclideanMFGPFitterBaseTestClass, BaseTestClass):
class EuclideanMFGPFitterTestCase(EuclideanMFGPFitterBaseTestClass):
  """ Unit tests for the MFGPFitter class. """
  # pylint: disable=too-many-locals

  def test_set_up(self):
    """ Tests if everything has been set up properly. """
    for ds in self.datasets:
      options = _get_mf_gp_options(ds.tune_noise)
      fitter = euclidean_gp.EuclideanMFGPFitter(ds.Z_tr, ds.X_tr, ds.Y_tr, options)
      # The number of hyperparameters should be 1 for the kernel, the total dimensionality
      # of the fidelity space and domain plus one more if we are tuning options.
      num_hps = 1 + ds.Z_tr.shape[1] + ds.X_tr.shape[1] + ds.tune_noise
      constructed_hp_bounds = ([fitter.scale_log_bounds] +
                               fitter.fidel_bandwidth_log_bounds +
                               fitter.domain_bandwidth_log_bounds)
      if ds.tune_noise:
        constructed_hp_bounds = [fitter.noise_var_log_bounds] + constructed_hp_bounds
      constructed_hp_bounds = np.array(constructed_hp_bounds)
      assert fitter.cts_hp_bounds.shape == (num_hps, 2)
      assert np.linalg.norm(constructed_hp_bounds - fitter.cts_hp_bounds) < 1e-5

  def test_marginal_likelihood(self):
    """ Test for marginal likelihood. """
    self.report(('Likelihood & Error for fitted MFGP. Probabilistic test, might fail.' +
                 ' The domain bandwidth should be smaller than the fidelity bandwidth ' +
                 'for the second dataset.'))
    self._test_marginal_likelihood(build_mfgp_with_dataset, fit_mfgp_with_dataset,
                                   'naive', 'fitted')

  def test_eval(self):
    """ Test for prediction. """
    self.report('Prediction for fitted Simple vs MF GPs. Probabilistic test, might fail.')
    num_successes = 0
    num_experiments = 0
    for ds_idx, ds in enumerate(self.datasets):
      for kernel_type in ['se', 'matern']:
        simple_gp = fit_simple_gp_with_dataset(ds, kernel_type)
        simple_preds, _ = simple_gp.eval(ds.ZX_te)
        simple_err = compute_average_sq_prediction_error(ds.Y_te, simple_preds)
        mfgp = fit_mfgp_with_dataset(ds, kernel_type)
        mf_preds, _ = mfgp.eval_at_fidel(ds.Z_te, ds.X_te)
        mf_err = compute_average_sq_prediction_error(ds.Y_te, mf_preds)
        # Book keep and report
        success = abs(mf_err - simple_err) < 1e-2
        num_successes += success
        num_experiments += 1
        self.report(('%d. (N,DZ,DX)= %s, kern=%s, simple-err=%0.4f, mfgp-err=%0.4f, ' +
          'succ=%d.')%(ds_idx+1, str(ds.Z_tr.shape + (ds.X_tr.shape[1],)), kernel_type,
          simple_err, mf_err, success), 'test_result')
    assert num_successes > 0.6 * num_experiments

  def test_draw_samples(self):
    """ Test for drawing samples. """
    self.report('Test for drawing samples. Probabilistic test, might fail.')
    total_coverage = 0
    num_test_pts = 100
    num_samples = 5 # Draw 5 samples at each point - just for testing.
    mfgp_instances = gen_mf_gp_instances()
    for inst in mfgp_instances:
      Z_test = np.random.random((num_test_pts, inst.fidel_dim))
      X_test = np.random.random((num_test_pts, inst.domain_dim))
      F_test = inst.post_gp.draw_mf_samples(num_samples, Z_test, X_test)
      post_mean, post_std = inst.post_gp.eval_at_fidel(Z_test, X_test, uncert_form='std')
      conf_band_width = 1.96
      ucb = post_mean + conf_band_width * post_std
      lcb = post_mean - conf_band_width * post_std
      below_ucb = F_test <= ucb
      above_lcb = F_test >= lcb
      coverage = (below_ucb * above_lcb).mean()
      total_coverage += coverage
      self.report(('(n, DZ, DX) = (%d, %d, %d)::: Coverage for 0.95 credible interval: ' +
                   '%0.4f')%(inst.num_data, inst.fidel_dim, inst.domain_dim, coverage),
                  'test_result')
    avg_coverage = total_coverage / len(mfgp_instances)
    avg_coverage_is_good = avg_coverage > 0.9
    self.report('Avg coverage (%0.3f) is larger than 0.9? %d'%(avg_coverage,
                avg_coverage_is_good), 'test_result')
    assert avg_coverage_is_good


# A Test for EuclideanMFGP (with additive models) -----------------------------
# class AddEuclideanMFGPFitterTestCase(EuclideanMFGPFitterBaseTestClass, BaseTestClass):
class AddEuclideanMFGPFitterTestCase(EuclideanMFGPFitterBaseTestClass):
  """ Unit tests for the MFGPFitter class. """

  def setUp(self):
    """ Set up for the tests. """
    self.datasets = gen_mf_gp_test_data()
    self.datasets = [self.datasets[i] for i in [0]] # dataset 1 is 1D so no real gains.

  def test_marginal_likelihood(self):
    """ Test for marginal likelihood. """
    self.report(('Likelihood & Error for non-additive vs additive MFGP. ' +
                 'Probabilistic test, might fail.'))
    self._test_marginal_likelihood(fit_mfgp_with_dataset, fit_add_mfgp_with_dataset,
                                   'non-add', 'additive')

# A Test for ExpDecay Kernel -----------------------------
class ExpDecayMFGPFitterTestCase(EuclideanMFGPFitterBaseTestClass, BaseTestClass):
  """ Unit tests for the MFGPFitter class. """

  def setUp(self):
    """ Set up for the tests. """
    self.datasets = gen_mf_gp_test_data()

  def test_marginal_likelihood(self):
    """ Test for marginal likelihood. """
    self.report(('Likelihood & Error for se and exp-decay MFGP. ' +
                 'Probabilistic test, might fail.'))
    self._test_marginal_likelihood(fit_mfgp_with_dataset,
                                   fit_mfgp_with_expdecay_fidel_kernel,
                                   'se', 'expdecay', compare_fits=False)



if __name__ == '__main__':
  execute_tests()




"""
  Unit tests for Multi-fidelity GP Bandits on Cartesian Product domains.
  -- kandasamy@cs.cmu.edu
"""
# Writing it as a separate file (unlike unittest_cp_random_optimiser.py) since these tests
# can be quite expensive.

import unittest

# Local imports
from . import gp_bandit
from .unittest_cp_random_optimiser import MFCPOptimiserBaseTestCase
from ..utils.base_test_class import BaseTestClass, execute_tests


class MFCPGPBanditTestCaseDefinitions(object):
  """ Definitions for unit tests for GP Bandits on cartesian product spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager, options, reporter):
    """ Instantiate the optimiser. """
    return gp_bandit.CPGPBandit(func_caller, worker_manager, options=options,
                                reporter=reporter)

  @classmethod
  def _run_optimiser(cls, prob_funcs, domain_config_file, worker_manager, max_capital,
                     mode, *args, **kwargs):
    """ Run the optimiser. """
    return gp_bandit.mf_cp_gpb_from_raw_args(prob_funcs[0], prob_funcs[1],
             domain_config_file, worker_manager=worker_manager, max_capital=max_capital,
             is_mf=True, mode=mode, *args, **kwargs)


@unittest.skip
class MFCPGPBanditTestCase(MFCPGPBanditTestCaseDefinitions,
                           MFCPOptimiserBaseTestCase,
                           BaseTestClass):
  """ Unit tests for GP Bandits on cartesian product spaces. """
  pass


if __name__ == '__main__':
  execute_tests()

